import configparser
import contextlib
import getpass
import importlib
import os
import pty
import re
import select
import shlex
import socket
import subprocess
import sys
import textwrap
import threading
import time
import warnings
from pathlib import Path
from typing import Dict, Optional, Union

try:
    import boto3
    import paramiko
    import sagemaker
    from sagemaker.estimator import EstimatorBase
    from sagemaker.mxnet import MXNet
    from sagemaker.pytorch import PyTorch
    from sagemaker.tensorflow import TensorFlow
    from sagemaker.xgboost import XGBoost
except ImportError:
    pass

from runhouse.constants import LOCAL_HOSTS

from runhouse.globals import configs, rns_client

from runhouse.logger import logger
from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.hardware.utils import ServerConnectionType
from runhouse.rns.utils.api import is_jsonable, relative_ssh_path, resolve_absolute_path
from runhouse.rns.utils.names import _generate_default_name


####################################################################################################
# Caching mechanisms for SSHTunnelForwarder
####################################################################################################

ssh_tunnel_cache = {}


def get_open_ssh_tunnel(address: str, ssh_port: int) -> Optional["SSHTunnelForwarder"]:
    if (address, ssh_port) in ssh_tunnel_cache:

        ssh_tunnel = ssh_tunnel_cache[(address, ssh_port)]
        # Initializes tunnel_is_up dictionary
        ssh_tunnel.check_tunnels()

        if (
            ssh_tunnel.is_active
            and ssh_tunnel.tunnel_is_up[ssh_tunnel.local_bind_address]
        ):
            return ssh_tunnel

        else:
            # If the tunnel is no longer active or up, pop it from the global cache
            ssh_tunnel_cache.pop((address, ssh_port))
    else:
        return None


def cache_open_ssh_tunnel(
    address: str,
    ssh_port: str,
    ssh_tunnel: "SSHTunnelForwarder",
):
    ssh_tunnel_cache[(address, ssh_port)] = ssh_tunnel


class SageMakerCluster(Cluster):
    DEFAULT_SERVER_HOST = "localhost"
    DEFAULT_INSTANCE_TYPE = "ml.m5.large"
    DEFAULT_REGION = "us-east-1"
    DEFAULT_USER = "root"
    # https://github.com/aws/deep-learning-containers/blob/master/available_images.md
    BASE_ECR_URL = "763104351884.dkr.ecr.us-east-1.amazonaws.com"

    # Default path for any estimator source code copied onto the cluster
    ESTIMATOR_SRC_CODE_PATH = "/opt/ml/code"
    ESTIMATOR_LOG_FILE = "sm_cluster.out"
    SSH_KEY_FILE_NAME = "sagemaker-ssh-gw"

    DEFAULT_SSH_PORT = 11022
    DEFAULT_CONNECTION_WAIT_TIME = 60  # seconds

    def __init__(
        self,
        name: str,
        role: str = None,
        profile: str = None,
        region: str = None,
        ssh_key_path: str = None,
        instance_id: str = None,
        instance_type: str = None,
        num_instances: int = None,
        image_uri: str = None,
        autostop_mins: int = None,
        connection_wait_time: int = None,
        estimator: Union["EstimatorBase", Dict] = None,
        job_name: str = None,
        server_host: str = None,
        server_port: int = None,
        domain: str = None,
        server_connection_type: str = None,
        ssl_keyfile: str = None,
        ssl_certfile: str = None,
        den_auth: bool = False,
        dryrun=False,
        **kwargs,
    ):
        """
        The Runhouse SageMaker cluster abstraction. This is where you can use SageMaker as a compute backend, just as
        you would an on-demand cluster (i.e. cloud VMs) or a BYO (i.e. on-prem) cluster. Additionally supports running
        dedicated training jobs using SageMaker Estimators.

        .. note::
            To build a cluster, please use the factory method :func:`sagemaker_cluster`.
        """
        super().__init__(
            name=name,
            ssh_creds=kwargs.pop("ssh_creds", {}),
            ssh_port=kwargs.pop("ssh_port", self.DEFAULT_SSH_PORT),
            server_host=server_host,
            server_port=server_port,
            server_connection_type=server_connection_type,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
            domain=domain,
            den_auth=den_auth,
            dryrun=dryrun,
            **kwargs,
        )
        self._connection_wait_time = connection_wait_time
        self._instance_type = instance_type
        self._num_instances = num_instances
        self._ssh_key_path = ssh_key_path

        # SSHEstimatorWrapper to facilitate the SSH connection to the cluster
        self._ssh_wrapper = None

        # Note: Relevant only if an estimator is explicitly provided
        self._estimator_entry_point = kwargs.get("estimator_entry_point")
        self._estimator_source_dir = kwargs.get("estimator_source_dir")
        self._estimator_framework = kwargs.get("estimator_framework")

        self.job_name = job_name

        # Set initial region - may be overwritten depending on the profile used
        self.region = region or self.DEFAULT_REGION

        # Set a default sessions initially - may overwrite depending on the profile loaded below
        self._set_boto_session()
        self._set_sagemaker_session()

        # Either use the user-provided instance_id, or look it up from the job_name
        self.instance_id = instance_id or (
            self._cluster_instance_id() if self.job_name else None
        )

        self._autostop_mins = (
            autostop_mins
            if autostop_mins is not None
            else configs.get("default_autostop")
        )

        self.estimator = self._load_estimator(estimator)

        self.role, self.profile = (
            self._load_role_and_profile(role, profile)
            if not dryrun
            else (role, profile)
        )
        logger.info(
            f"Using SageMaker execution role: `{self.role}` and profile: `{self.profile}`"
        )

        self.image_uri = self._load_image_uri(image_uri)

        # Note: Setting instance ID as cluster IP for compatibility with Cluster parent class methods
        self.address = self.instance_id

    def config(self, condensed=True):
        config = super().config(condensed)
        config.update(
            {
                "instance_id": self.instance_id,
                "role": self.role,
                "region": self.region,
                "profile": self.profile,
                "ssh_key_path": self.ssh_key_path,
                "job_name": self.job_name,
                "instance_type": self.instance_type,
                "num_instances": self.num_instances,
                "image_uri": self.image_uri,
                "autostop_mins": self._autostop_mins,
                "connection_wait_time": self.connection_wait_time,
            }
        )

        # If running a dedicated job on the cluster, add the estimator config
        if self.estimator and (
            self._estimator_source_dir and self._estimator_entry_point
        ):
            config.update(
                {
                    "estimator_entry_point": self._estimator_entry_point,
                    "estimator_source_dir": str(self._estimator_source_dir),
                }
            )

            if isinstance(self.estimator, EstimatorBase):
                # Serialize the estimator before saving it down in the config
                selected_attrs = {
                    key: value
                    for key, value in self.estimator.__dict__.items()
                    if is_jsonable(value)
                }
                # Estimator types: mxnet, tensorflow, keras, pytorch, onnx, xgboost
                self._estimator_framework = type(self.estimator).__name__
                config.update(
                    {
                        "estimator": selected_attrs,
                        "estimator_framework": self._estimator_framework,
                    }
                )

            if isinstance(self.estimator, dict):
                config.update({"estimator": self.estimator})

        return config

    @property
    def hosts_path(self):
        return Path("~/.ssh/known_hosts").expanduser()

    @property
    def ssh_config_file(self):
        return Path("~/.ssh/config").expanduser()

    @property
    def ssh_key_path(self):
        """Relative path to the private SSH key used to connect to the cluster."""
        if self._ssh_key_path:
            return relative_ssh_path(self._ssh_key_path)

        # Default relative path
        return f"~/.ssh/{self.SSH_KEY_FILE_NAME}"

    @ssh_key_path.setter
    def ssh_key_path(self, ssh_key_path):
        self._ssh_key_path = ssh_key_path

    @property
    def num_instances(self):
        if self._num_instances:
            return self._num_instances
        elif self.estimator:
            return self.estimator.instance_count
        else:
            return 1

    @num_instances.setter
    def num_instances(self, num_instances):
        self._num_instances = num_instances

    @property
    def connection_wait_time(self):
        """Amount of time the SSH helper will wait inside SageMaker before it continues normal execution"""
        if self._connection_wait_time is not None:
            return self._connection_wait_time
        elif self.estimator and (
            self._estimator_source_dir and self._estimator_entry_point
        ):
            # Allow for connecting to the instance before the job starts (e.g. training)
            return self.DEFAULT_CONNECTION_WAIT_TIME
        else:
            # For inference and others, always up and running
            return 0

    @connection_wait_time.setter
    def connection_wait_time(self, connection_wait_time):
        self._connection_wait_time = connection_wait_time

    @property
    def instance_type(self):
        if self._instance_type:
            return self._instance_type
        elif self.estimator:
            return self.estimator.instance_type
        else:
            return self.DEFAULT_INSTANCE_TYPE

    @instance_type.setter
    def instance_type(self, instance_type):
        self._instance_type = instance_type

    @property
    def default_bucket(self):
        """Default bucket to use for storing the cluster's authorized public keys."""
        return self._sagemaker_session.default_bucket()

    @property
    def _use_https(self) -> bool:
        # Note: Since always connecting via SSM no need for HTTPS
        return False

    @property
    def _extra_ssh_args(self):
        """Extra SSH arguments to be used when connecting to the cluster."""
        # Note - port 12345 can be used for Python Debug Server: "-R localhost:12345:localhost:12345"
        # https://github.com/aws-samples/sagemaker-ssh-helper#remote-debugging-with-pycharm-debug-server-over-ssh
        return f"-L localhost:{self.ssh_port}:localhost:22"

    @property
    def _s3_keys_path(self):
        """Path to public key stored for the cluster on S3. When initializing the cluster, the public key
        is copied by default to an authorized keys file in this location."""
        return f"s3://{self.default_bucket}/ssh-authorized-keys/"

    @property
    def _ssh_public_key_path(self):
        return f"{self._abs_ssh_key_path}.pub"

    @property
    def _abs_ssh_key_path(self):
        return resolve_absolute_path(self.ssh_key_path)

    @property
    def _ssh_key_comment(self):
        """Username and hostname to be used as the comment for the public key."""
        return f"{getpass.getuser()}@{socket.gethostname()}"

    @property
    def _s3_client(self):
        if self._boto_session is None:
            self._set_boto_session()

        return self._boto_session.client("s3")

    def _get_env_activate_cmd(self, env=None):
        """Prefix for commands run on the cluster. Ensure we are running all commands in the conda environment
        and not the system default python."""
        # TODO [JL] Can SageMaker handle this for us?
        if env:
            from runhouse.resources.envs import _get_env_from

            return _get_env_from(env)._activate_cmd
        return "source /opt/conda/bin/activate"

    def _set_boto_session(self, profile_name: str = None):
        self._boto_session = boto3.Session(
            region_name=self.region, profile_name=profile_name
        )

    def _set_sagemaker_session(self):
        """Create a SageMaker session required for using the SageMaker APIs."""
        self._sagemaker_session = sagemaker.Session(boto_session=self._boto_session)

    def set_connection_defaults(self):
        if (
            "aws-cli/2."
            not in subprocess.run(
                ["aws", "--version"], capture_output=True, text=True
            ).stdout
        ):
            raise RuntimeError(
                "SageMaker SDK requires AWS CLI v2. You may also need to run `pip uninstall awscli` to ensure "
                "the right version is being used. For more info: https://www.run.house/docs/api/python/cluster#id9"
            )

        if self.ssh_key_path:
            self.ssh_key_path = relative_ssh_path(self.ssh_key_path)
        else:
            self.ssh_key_path = None

        if (
            self.server_connection_type is not None
            and self.server_connection_type != ServerConnectionType.AWS_SSM
        ):
            raise ValueError(
                "SageMaker Cluster currently requires a server connection type of `aws_ssm`."
            )
        self.server_connection_type = ServerConnectionType.AWS_SSM.value

        if self.server_host and self.server_host not in LOCAL_HOSTS:
            raise ValueError(
                "SageMaker Cluster currently requires a server host of `localhost` or `127.0.0.1`"
            )

    # -------------------------------------------------------
    # Cluster State & Lifecycle Methods
    # -------------------------------------------------------
    def restart_server(
        self,
        _rh_install_url: str = None,
        resync_rh: bool = True,
        restart_ray: bool = True,
        env: Union[str, "Env"] = None,
        restart_proxy: bool = False,
    ):
        """Restart the RPC server on the SageMaker instance.

        Args:
            resync_rh (bool): Whether to resync runhouse. (Default: ``True``)
            restart_ray (bool): Whether to restart Ray. (Default: ``True``)
            env (str or Env): Env to restart the server from. If not provided
                will use default env on the cluster.
            restart_proxy (bool): Whether to restart nginx on the cluster, if configured. (Default: ``False``)
        Example:
            >>> rh.sagemaker_cluster("sagemaker-cluster").restart_server()
        """
        return super().restart_server(
            _rh_install_url, resync_rh, restart_ray, env, restart_proxy
        )

    def check_server(self, restart_server=True):
        if self.on_this_cluster():
            return

        if not self.instance_id or not self.is_up():
            logger.info(f"Cluster {self.name} is not up, bringing it up now.")
            self.up_if_not()

        if not self._http_client:
            try:
                self.connect_server_client()
                logger.info(
                    f"Checking server {self.name} with instance ID: {self.instance_id}"
                )

                self.client.check_server()
                logger.info(f"Server {self.instance_id} is up.")
            except:
                if restart_server:
                    logger.info(
                        f"Server {self.instance_id} is up, but the API server may not be up."
                    )
                    self.run(
                        [
                            "sudo apt-get install screen -y "
                            "&& sudo apt-get install rsync -y"
                        ]
                    )
                    # Restart the server inside the base conda env
                    self.restart_server(
                        resync_rh=True,
                        restart_ray=True,
                        env=None,
                    )
                    logger.info(f"Checking server {self.instance_id} again.")

                    self.client.check_server()
                else:
                    raise ValueError(
                        f"Could not connect to SageMaker instance {self.instance_id}"
                    )

    def up(self):
        """Up the cluster.

        Example:
            >>> rh.sagemaker_cluster("sagemaker-cluster").up()
        """
        logger.info("Preparing to launch a new SageMaker cluster")
        self._launch_new_cluster()

        return self

    def up_if_not(self):
        """Bring up the cluster if it is not up. No-op if cluster is already up.

        Example:
            >>> rh.sagemaker_cluster("sagemaker-cluster").up_if_not()
        """
        if not self.is_up():
            self.address = None
            self.job_name = None
            self.instance_id = None
            self.up()
        return self

    def is_up(self) -> bool:
        """Check if the cluster is up.

        Example:
            >>> rh.sagemaker_cluster("sagemaker-cluster").is_up()
        """
        try:
            resp: dict = self.status()
            status = resp.get("TrainingJobStatus")
            # Up if the instance is in progress
            return status == "InProgress"
        except:
            return False

    def teardown(self):
        """Teardown the SageMaker instance.

        Example:
            >>> rh.sagemaker_cluster(name="sagemaker-cluster").teardown()
        """
        self._stop_instance(delete_configs=False)

    def teardown_and_delete(self):
        """Teardown the SageMaker instance and delete from RNS configs.

        Example:
            >>> rh.sagemaker_cluster(name="sagemaker-cluster").teardown_and_delete()
        """
        self._stop_instance()

    def keep_warm(self, autostop_mins: int = -1):
        """Keep the cluster warm for given number of minutes after inactivity.

        Args:
            autostop_mins (int): Amount of time (in minutes) to keep the cluster warm after inactivity.
                If set to ``-1``, keep cluster warm indefinitely. (Default: ``-1``)
        """
        self._update_autostop(autostop_mins)

        return self

    def __getstate__(self):
        """Delete non-serializable elements (e.g. sagemaker session object) before pickling."""
        state = self.__dict__.copy()
        state["_sagemaker_session"] = None
        state["_http_client"] = None
        state["_rpc_tunnel"] = None
        return state

    @contextlib.contextmanager
    def pause_autostop(self):
        """Context manager to temporarily pause autostop."""
        self._update_autostop(autostop_mins=-1)
        yield
        self._update_autostop(self._autostop_mins)

    def status(self) -> dict:
        """
        Get status of SageMaker cluster.

        Example:
            >>> status = rh.sagemaker_cluster("sagemaker-cluster").status()
        """
        try:
            return self._sagemaker_session.describe_training_job(self.job_name)
        except:
            return {}

    # -------------------------------------------------------
    # SSH APIs
    # -------------------------------------------------------
    def ssh_tunnel(
        self, local_port, remote_port=None, num_ports_to_try: int = 0, retry=True
    ) -> "SSHTunnelForwarder":
        from sshtunnel import BaseSSHTunnelForwarderError, SSHTunnelForwarder

        tunnel = get_open_ssh_tunnel(self.address, self.ssh_port)
        if tunnel and tunnel.local_bind_port == local_port:
            logger.info(
                f"SSH tunnel on ports {local_port, remote_port} already created with the cluster"
            )
            return tunnel

        try:
            remote_bind_addresses = ("127.0.0.1", local_port)
            local_bind_addresses = ("", local_port)

            ssh_tunnel = SSHTunnelForwarder(
                self.DEFAULT_SERVER_HOST,
                ssh_username=self.DEFAULT_USER,
                ssh_pkey=self._abs_ssh_key_path,
                ssh_port=self.ssh_port,
                remote_bind_address=remote_bind_addresses,
                local_bind_address=local_bind_addresses,
                set_keepalive=1800,
            )

            # Start the SSH tunnel
            ssh_tunnel.start()

            # Update the SSH config for the cluster with the connected SSH port
            self._add_or_update_ssh_config_entry()

            logger.info("SSH connection has been successfully created with the cluster")

        except BaseSSHTunnelForwarderError as e:
            if not retry:
                # Failed to create the SSH tunnel object even after successfully refreshing the SSM session
                raise BaseSSHTunnelForwarderError(
                    f"{e} Make sure ports {self.server_port} and {self.ssh_port} are "
                    f"not already in use."
                )

            # Refresh the SSM session, which should bind the HTTP and SSH ports to localhost which are forwarded
            # to the cluster
            self._refresh_ssm_session_with_cluster(num_ports_to_try)

            # Retry creating the SSH tunnel once the session has been refreshed
            return self.ssh_tunnel(
                local_port, remote_port, num_ports_to_try, retry=False
            )

        cache_open_ssh_tunnel(self.address, self.ssh_port, ssh_tunnel)
        return ssh_tunnel

    def ssh(self, interactive: bool = True):
        """SSH into the cluster.

        Args:
            interactive (bool): Whether to start an interactive shell or not (Default: ``True``).

        Example:
            >>> rh.sagemaker_cluster(name="sagemaker-cluster").ssh()
        """

        if (self.address, self.ssh_port) not in ssh_tunnel_cache:
            # Make sure SSM session and SSH tunnels are up before running the command
            self.connect_server_client()

        if not interactive:
            logger.info(
                f"Created SSH tunnel with the cluster. To SSH into the cluster, run: `ssh {self.name}`"
            )
            return

        head_fd, worker_fd = pty.openpty()
        ssh_process = subprocess.Popen(
            ["ssh", "-o", "StrictHostKeyChecking=no", self.name],
            stdin=worker_fd,
            stdout=worker_fd,
            stderr=worker_fd,
            universal_newlines=True,
        )

        # Close the worker_fd in the parent process as it's not needed there
        os.close(worker_fd)

        # Wait for the SSH process to initialize
        select.select([head_fd], [], [])

        # Interact with the SSH process through the head_fd
        try:
            while True:
                if head_fd in select.select([head_fd], [], [], 0)[0]:
                    output = os.read(head_fd, 1024).decode()
                    print(output, end="")

                if sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
                    user_input = sys.stdin.readline()
                    try:
                        os.write(head_fd, user_input.encode())
                    except OSError:
                        pass

                    # terminate the SSH process gracefully
                    if user_input.strip() == "exit":
                        break
        except Exception as e:
            raise e
        finally:
            # Close the head_fd and terminate the SSH process when done
            os.close(head_fd)
            ssh_process.terminate()

    def _run_commands_with_ssh(
        self,
        commands: list,
        cmd_prefix: str,
        stream_logs: bool,
        node: str = None,
        port_forward: int = None,
        require_outputs: bool = True,
    ):
        return_codes = []
        for command in commands:
            if command.startswith("rsync"):
                try:
                    result = subprocess.run(
                        command,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        text=True,
                    )
                    return_codes.append(
                        (result.returncode, result.stdout, result.stderr)
                    )
                except subprocess.CalledProcessError as e:
                    return_codes.append((255, "", str(e)))
            else:
                # Host can be replaced with name (as reflected in the ~/.ssh/config file)
                from runhouse.resources.hardware.sky_ssh_runner import SkySSHRunner

                runner = SkySSHRunner(
                    self.name,
                    port=self.ssh_port,
                    ssh_user=self.DEFAULT_USER,
                    ssh_private_key=self._abs_ssh_key_path,
                    ssh_control_name=f"{self.name}:{self.ssh_port}",
                )
                command = f"{cmd_prefix} {command}" if cmd_prefix else command
                logger.debug(f"Running command on {self.name}: {command}")
                return_code, stdout, stderr = runner.run(
                    command,
                    require_outputs=require_outputs,
                    stream_logs=stream_logs,
                    port_forward=port_forward,
                )

                if (
                    return_code != 0
                    and "dpkg: error processing package install-info" in stdout
                ):
                    # **NOTE**: there may be issues with some SageMaker GPUs post installation script which
                    # leads to an error which looks something like: "installed install-info package post-installation
                    # script subprocess returned error exit status 2"
                    # https://askubuntu.com/questions/1034961/cant-upgrade-error-etc-environment-source-not-found-and-error-processin
                    # /etc/environment file may also be corrupt, replacing with an empty file allows
                    # subsequent python commands to run
                    self._run_commands_with_ssh(
                        commands=[
                            "cd /var/lib/dpkg/info && sudo rm *.postinst "
                            "&& sudo mv /etc/environment /etc/environment_broken "
                            "&& sudo touch /etc/environment "
                            f"&& {command}"
                        ],
                        cmd_prefix=cmd_prefix,
                        stream_logs=stream_logs,
                    )

                return_codes.append((return_code, stdout, stderr))

        return return_codes

    # -------------------------------------------------------
    # Cluster Provisioning & Launching
    # -------------------------------------------------------
    def _create_launch_estimator(self):
        """Create the estimator object used for launching the cluster. If a custom estimator is provided, use that.
        Otherwise, use a Runhouse default estimator to launch.
        **Note If an estimator is provided, Runhouse will override the entry point and source dir to use the
        default Runhouse entry point and source dir. This is to ensure that the connection to the cluster
        can be maintained even if the job fails, after it has completed, or if autostop is enabled.**"""
        # Note: these entry points must point to the existing local files
        default_entry_point = "launch_instance.py"
        full_module_name = f"resources/hardware/sagemaker/{default_entry_point}"

        entry_point_path = self._get_path_for_module(full_module_name)
        source_dir_path = os.path.dirname(entry_point_path)

        # Set default_entry_point and default_source_dir
        default_source_dir = source_dir_path

        if self.estimator:
            # Save the original entry point and source dir to be used on the cluster for running the estimator
            self._estimator_entry_point = self.estimator.entry_point
            self._estimator_source_dir = self.estimator.source_dir

            # Update the estimator with the Runhouse custom entry point and source dir
            # When the job is initialized, it will run through the Runhouse entry point, which will manage the
            # running of the custom estimator
            self.estimator.entry_point = default_entry_point
            self.estimator.source_dir = default_source_dir

            return self.estimator

        else:
            # No estimator provided, use the Runhouse custom estimator (using PyTorch by default)
            estimator_dict = {
                "instance_count": self.num_instances,
                "role": self.role,
                "image_uri": self.image_uri,
                "framework_version": "2.0.1",
                "py_version": "py310",
                "entry_point": default_entry_point,
                "source_dir": default_source_dir,
                "instance_type": self.instance_type,
                # https://docs.aws.amazon.com/sagemaker/latest/dg/train-warm-pools.html
                "keep_alive_period_in_seconds": 3600,
            }

            return PyTorch(**estimator_dict)

    def _launch_new_cluster(self):
        self.estimator = self._create_launch_estimator()

        logger.info(
            f"Launching a new SageMaker cluster (num instances={self.num_instances}) on instance "
            f"type: {self.instance_type}"
        )

        self._create_new_instance()

        # If no name provided, use the autogenerated name
        self.job_name = self.estimator.latest_training_job.name

        self.instance_id = self._cluster_instance_id()

        # For compatibility with parent Cluster class methods which use an address
        self.address = self.instance_id

        logger.info(f"New SageMaker instance started with ID: {self.instance_id}")

        # Remove stale entries from the known hosts file
        self._filter_known_hosts()

        logger.info("Creating session with cluster via SSM")
        self._create_ssm_session_with_cluster()

        self.check_server()

        if self._estimator_source_dir and self._estimator_entry_point:
            # Copy the provided estimator's code to the cluster - Runhouse will then manage running the job in order
            # to preserve control over the cluster's autostop
            self._sync_estimator_to_cluster()
            logger.info(
                f"Logs for the estimator can be viewed on the cluster in "
                f"path: {self.ESTIMATOR_SRC_CODE_PATH}/{self.ESTIMATOR_LOG_FILE}"
            )

        logger.info(
            f"Connection with {self.name} has been created. You can SSH onto "
            f"the cluster with the CLI using: ``ssh {self.name}``"
        )

    def _create_new_instance(self):
        from sagemaker_ssh_helper.wrapper import SSHEstimatorWrapper

        # Make sure the SSHEstimatorWrapper is being used by the estimator, this is necessary for
        # enabling the SSH tunnel to the cluster
        # https://github.com/aws-samples/sagemaker-ssh-helper/tree/main#step-2-modify-your-start-training-job-code
        ssh_dependency_dir = SSHEstimatorWrapper.dependency_dir()
        if ssh_dependency_dir not in self.estimator.dependencies:
            self.estimator.dependencies.append(ssh_dependency_dir)

        # Create the SSH wrapper & run the job
        self._ssh_wrapper = SSHEstimatorWrapper.create(
            self.estimator, connection_wait_time_seconds=self.connection_wait_time
        )

        self._start_instance()

    def _create_ssm_session_with_cluster(self, num_ports_to_try: int = 5):
        """Create a session with the cluster. Runs a bash script containing a series of commands which use existing
        SSH keys or generate new ones needed to authorize the connection with the cluster via the AWS SSM.
        These commands are run when the cluster is initially provisioned, or for subsequent connections if the session
        is longer active. Once finished, the SSH port and HTTP port will be bound to processes on
        localhost which are forwarded to the cluster."""
        # https://github.com/aws-samples/sagemaker-ssh-helper/tree/main#forwarding-tcp-ports-over-ssh-tunnel
        base_command = self._load_base_command_for_ssm_session()

        connected = False
        while not connected:
            command = f"{base_command} {self._extra_ssh_args}"

            try:
                if num_ports_to_try == 0:
                    raise ConnectionError(
                        f"Failed to create SSM session and connect to {self.name} after repeated attempts."
                        f"Make sure SSH keys exist in local path: {self._abs_ssh_key_path}"
                    )

                logger.debug(f"Running command: {command}")

                # Define an event to signal completion of the SSH tunnel setup
                tunnel_setup_complete = threading.Event()

                # Manually allocate a pseudo-terminal to prevent a "pseudo-terminal not allocated" error
                head_fd, worker_fd = pty.openpty()

                def run_ssm_session_cmd():
                    # Execute the command with the pseudo-terminal in a separate thread
                    process = subprocess.Popen(
                        command,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stdin=subprocess.PIPE,
                    )

                    # Close the worker file descriptor as we don't need it
                    os.close(worker_fd)

                    # Close the master file descriptor after reading the output
                    os.close(head_fd)

                    # Wait for the process to complete and collect its return code
                    process.wait()

                    # Signal that the tunnel setup is complete
                    tunnel_setup_complete.set()

                tunnel_thread = threading.Thread(target=run_ssm_session_cmd)
                tunnel_thread.daemon = True  # Set the thread as a daemon, so it won't block the main thread

                # Start the SSH tunnel thread
                tunnel_thread.start()

                # Give time for the SSM session to start, SSH keys to be copied onto the cluster, and the SSH port
                # forwarding command to run
                tunnel_setup_complete.wait(timeout=30)

                if not self._ports_are_in_use():
                    # Command should bind SSH port and HTTP port on localhost, if this is not the case try re-running
                    # the bash script with a different set of ports
                    # E.g. ❯ lsof -i:11022,32300
                    # COMMAND   PID   USER   FD   TYPE             DEVICE SIZE/OFF NODE NAME
                    # ssh     97115 myuser    3u  IPv4 0xcf81f230786cc9fd      0t0  TCP localhost:32300 (LISTEN)
                    # ssh     97115 myuser    6u  IPv4 0xcf81f230786eff6d      0t0  TCP localhost:11022 (LISTEN)
                    raise ConnectionError

                # Update the SSH config for the cluster with the connected SSH port
                self._add_or_update_ssh_config_entry()

                connected = True

                logger.info(
                    f"Created SSM session using ports {self.ssh_port} and {self.server_port}. "
                    f"All active sessions can be viewed with: ``aws ssm describe-sessions --state Active``"
                )

            except (ConnectionError, subprocess.CalledProcessError):
                # Try re-running with updated the ports - possible the ports are already in use
                self.server_port += 1
                self.ssh_port += 1
                num_ports_to_try -= 1
                pass

    def _start_instance(self):
        """Call the SageMaker CreateTrainingJob API to start the training job on the cluster."""
        # TODO [JL] Note: Keeping private until re-running training jobs on the same cluster is supported
        if not self.estimator:
            logger.warning("No estimator found, cannot run job.")
            return

        # NOTE: underscores not allowed for training job name - must match: ^[a-zA-Z0-9](-*[a-zA-Z0-9]){0,62}
        self.estimator.fit(
            wait=False,
            job_name=self.job_name or _generate_default_name(self.name, sep="-"),
        )

    def _load_base_command_for_ssm_session(self) -> str:
        """Bash script command for creating the SSM session and uploading the SSH keys to the cluster. Will try
        reusing existing keys locally if they exist, otherwise will generate new ones locally and copy them to s3."""
        private_key_path = self._abs_ssh_key_path
        public_key_path = self._ssh_public_key_path

        resource_name = "resources/hardware/sagemaker/start-ssm-proxy-connection.sh"
        script_path = self._get_path_for_module(resource_name)

        os.chmod(script_path, 0o755)

        s3_key_path = self._s3_keys_path

        # bash script which creates an SSM session with the cluster
        base_command = (
            f'bash {script_path} "{self.instance_id}" '
            f'"{s3_key_path}" "{private_key_path}" "{self.region}"'
        )

        bucket = s3_key_path.split("/")[2]
        key = "/".join(s3_key_path.split("/")[3:])

        if Path(public_key_path).exists() and Path(private_key_path).exists():
            # If the key pair exists locally, make sure a matching public key also exists in s3
            with open(self._ssh_public_key_path, "r") as f:
                public_key = f.read()

            self._add_public_key_to_authorized_keys(bucket, key, public_key)

            return base_command

        # If no private + public keys exists generate a new key pair from scratch
        logger.warning(
            f"No private + public keypair found in local path: {private_key_path}. Generating a new key pair "
            "locally and uploading the new public key to s3"
        )
        self._create_new_ssh_key_pair(bucket, key)

        return base_command

    def _refresh_ssm_session_with_cluster(self, num_ports_to_try: int = 5):
        """Reconnect to the cluster via the AWS SSM. This bypasses the step of creating a new SSH key which was already
        done when upping the cluster. Note: this assumes the session has previously been created, which we do when
        the cluster has been upped.

        To view all sessions: ``aws ssm describe-sessions --state Active``
        """
        ssh_key_path = self._abs_ssh_key_path
        public_key_path = self._ssh_public_key_path

        if not Path(ssh_key_path).exists() and not Path(public_key_path).exists():
            logger.warning(
                f"SSH key pairs not found in paths: {ssh_key_path} and {public_key_path}"
            )
            self._create_ssm_session_with_cluster()

        # https://github.com/aws-samples/sagemaker-ssh-helper/blob/main/sagemaker_ssh_helper/sm-connect-ssh-proxy
        full_module_name = "resources/hardware/sagemaker/refresh-ssm-session.sh"
        script_path = self._get_path_for_module(full_module_name)

        os.chmod(script_path, 0o755)

        # Remove stale entries from the known hosts file - this is import for avoiding collisions when
        # subsequent clusters are created as the IP address is added to the file as: [localhost]:11022
        self._filter_known_hosts()

        num_attempts = num_ports_to_try
        connected = False

        while not connected:
            command = [
                script_path,
                self.instance_id,
                ssh_key_path,
                self.region,
            ] + shlex.split(self._extra_ssh_args)

            if num_ports_to_try == 0:
                raise ConnectionError(
                    f"Failed to create connection with {self.name} after {num_attempts} attempts "
                    f"(cluster status=`{self.status().get('TrainingJobStatus')}`). Make sure that another SageMaker "
                    f"cluster is not already active, that AWS CLI V2 is installed, and that the path has been properly "
                    f"added to your bash profile"
                    f"(https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-troubleshooting.html). "
                    f"If the error persists, try running the command to create the session "
                    f"manually: `bash {' '.join(command)}`"
                )

            try:
                subprocess.Popen(
                    command,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    stdin=subprocess.PIPE,
                )

                # Give enough time for the aws ssm + ssh port forwarding commands in the script to complete
                # Better to wait a few more seconds than to restart the HTTP server on the cluster unnecessarily
                time.sleep(8)

                if not self._ports_are_in_use():
                    # Command should bind SSH port and HTTP port on localhost, if this is not the case try re-running
                    # with different ports
                    raise socket.error

                connected = True
                logger.info("Successfully refreshed SSM session")

            except socket.error:
                # If the refresh didn't work try connecting with a different port - could be that the port
                # is already taken
                self.server_port += 1
                self.ssh_port += 1
                num_ports_to_try -= 1
                pass

    # -------------------------------------------------------
    # SSH Keys Management
    # -------------------------------------------------------
    def _create_new_ssh_key_pair(self, bucket, key):
        """Create a new private / public key pairing needed for SSHing into the cluster."""
        private_key_path = self._abs_ssh_key_path
        ssh_key = paramiko.RSAKey.generate(bits=2048)

        ssh_key.write_private_key_file(private_key_path)
        os.chmod(private_key_path, 0o600)

        # Set the comment for the public key to include username and hostname
        comment = self._ssh_key_comment
        public_key = f"ssh-rsa {ssh_key.get_base64()} {comment}"

        # Update the public key locally
        self._write_public_key(public_key)

        # Update the public key in s3
        self._add_public_key_to_authorized_keys(bucket, key, public_key)

    def _add_public_key_to_authorized_keys(
        self, bucket: str, key: str, public_key: str
    ) -> None:
        """Add the public key to the authorized keys file stored in S3. This file will get copied onto the cluster's
        authorized keys file."""
        path_to_auth_keys = self._path_to_auth_keys(key)
        authorized_keys: str = self._load_authorized_keys(bucket, path_to_auth_keys)

        if not authorized_keys:
            # Create a new authorized keys file
            logger.info(
                f"No authorized keys file found in s3 path: {self._s3_keys_path}. Creating and uploading a new file."
            )
            self._upload_key_to_s3(bucket, path_to_auth_keys, public_key)
            return

        if public_key not in authorized_keys:
            # Add the public key to the existing authorized keys saved in s3
            authorized_keys += f"\n{public_key}\n"
            logger.info(
                f"Adding public key to authorized keys file saved for the cluster "
                f"in path: {self._s3_keys_path}"
            )
            self._upload_key_to_s3(bucket, path_to_auth_keys, authorized_keys)

    def _path_to_auth_keys(self, key):
        """Path to the authorized keys file stored in the s3 bucket for the role ARN associated with the cluster."""
        return key + f"{self.SSH_KEY_FILE_NAME}.pub"

    def _write_public_key(self, public_key: str):
        """Update the public key stored locally."""
        with open(self._ssh_public_key_path, "w") as f:
            f.write(public_key)

    def _upload_key_to_s3(self, bucket, key, body):
        """Save a public key to the authorized file in the default bucket for given SageMaker role."""
        self._s3_client.put_object(
            Bucket=bucket,
            Key=key,
            Body=body,
        )

    def _load_authorized_keys(self, bucket, auth_keys_file) -> Union[str, None]:
        """Load the authorized keys file for this AWS role stored in S3. If no file exists, return None."""
        try:
            response = self._s3_client.get_object(Bucket=bucket, Key=auth_keys_file)
            existing_pub_keys = response["Body"].read().decode("utf-8")
            return existing_pub_keys

        except self._s3_client.exceptions.NoSuchKey:
            # No authorized keys file exists in s3 for this role
            return None

        except Exception as e:
            raise e

    # -------------------------------------------------------
    # Cluster Helpers
    # -------------------------------------------------------
    def _rsync(self, source: str, dest: str, up: bool, contents: bool = False):
        source = source + "/" if not source.endswith("/") else source
        dest = dest + "/" if not dest.endswith("/") else dest

        command = (
            f"rsync -rvh --exclude='.git' --exclude='venv*/' --exclude='dist/' --exclude='docs/' "
            f"--exclude='__pycache__/' --exclude='.*' "
            f"--include='.rh/' -e 'ssh -o StrictHostKeyChecking=no "
            f"-i {self._abs_ssh_key_path} -p {self.ssh_port}' {source} root@localhost:{dest}"
        )

        logger.info(f"Syncing {source} to: {dest} on cluster")
        return_codes = self.run([command])
        if return_codes[0][0] != 0:
            logger.error(f"rsync to SageMaker cluster failed: {return_codes[0][1]}")

    def _cluster_instance_id(self):
        """Get the instance ID of the cluster. This is the ID of the instance running the training job generated
        by SageMaker."""
        if self._ssh_wrapper:
            # This is a hack to effectively do list.get(0, None)
            return next(iter(self._ssh_wrapper.get_instance_ids()), None)

        from sagemaker_ssh_helper.manager import SSMManager

        ssm_manager = SSMManager(region_name=self.region)
        ssm_manager.redo_attempts = 0
        instance_ids = ssm_manager.get_training_instance_ids(self.job_name)
        # This is a hack to effectively do list.get(0, None)
        return next(iter(instance_ids), None)

    def _get_path_for_module(self, resource_name: str) -> str:
        import importlib.resources as pkg_resources

        package_name = "runhouse"
        script_path = str(pkg_resources.files(package_name) / resource_name)
        return script_path

    def _load_role_and_profile(self, role, profile):
        """Load the SageMaker role and profile used for launching and connecting to the cluster.
        Role can be provided as an ARN or a name. If provided, will search for the profile containing this role
        in local AWS configs. If no profile is provided, try loading from the environment variable ``AWS_PROFILE``,
        otherwise default to using the ``default`` profile."""
        if self.estimator:
            # If using an estimator must provide a name or full ARN
            # https://sagemaker.readthedocs.io/en/stable/api/training/estimators.html
            role = self.estimator.role

        if role and role.startswith("arn:aws"):
            profile = profile or self._load_profile_and_region_for_role(role)

        else:
            # https://docs.aws.amazon.com/sagemaker/latest/dg/sagemaker-roles.html
            profile = profile or os.environ.get("AWS_PROFILE")
            if profile is None:
                profile = "default"
                logger.warning(
                    f"No profile provided or environment variable set to `AWS_PROFILE`, using the {profile} profile"
                )

        try:
            # Update the sessions using the profile provided
            self._set_boto_session(profile_name=profile)
            self._set_sagemaker_session()

            # If no role explicitly provided, use sagemaker to get it via the profile
            role = role or sagemaker.get_execution_role(
                sagemaker_session=self._sagemaker_session
            )
        except Exception as e:
            if self.on_this_cluster():
                # If we're on the cluster, we may not have the profile or role saved locally, but should still be able
                # to create the cluster object (e.g. for rh.here).
                pass
            else:
                raise e

        return role, profile

    def _load_profile_and_region_for_role(self, role_arn: str) -> Union[str, None]:
        """Find the profile (and region) associated with a particular role ARN. If no profile is found, return None."""
        try:
            aws_dir = os.path.expanduser("~/.aws")
            credentials_path = os.path.join(aws_dir, "credentials")
            config_path = os.path.join(aws_dir, "config")

            profiles_with_role = set()

            for path in [credentials_path, config_path]:
                config = configparser.ConfigParser()
                config.read(path)

                for section in config.sections():
                    config_section = config[section]
                    config_role_arn = config_section.get("role_arn")
                    if config_role_arn == role_arn:
                        # Add just the name of the profile (not the full section heading)
                        profiles_with_role.add(section.split(" ")[-1])

                    # Update the region to use the one associated with this profile
                    profile_region = config_section.get("region")
                    if profile_region != self.region:
                        warnings.warn(
                            f"Updating region based on AWS config to: {profile_region}"
                        )
                        self.region = profile_region

            if not profiles_with_role:
                return None

            profiles = list(profiles_with_role)
            profile = profiles[0]

            if len(profiles) > 1:
                logger.warning(
                    f"Found multiple profiles associated with the same role. Using the first "
                    f"one ({profile})"
                )

            return profile

        except Exception as e:
            logger.warning(f"Could not find a profile for role {role_arn}: {e}")
            return None

    def _load_image_uri(self, image_uri: str = None) -> str:
        """Load the docker image URI used for launching the SageMaker instance. If no image URI is provided, use
        a default image based on the instance type."""
        if image_uri:
            return image_uri

        if self.estimator:
            return self.estimator.image_uri

        return self._base_image_uri()

    def _stop_instance(self, delete_configs=True):
        """Stop the SageMaker instance. Optionally remove its config from RNS."""
        self._sagemaker_session.stop_training_job(job_name=self.job_name)

        if self.is_up():
            raise Exception(f"Failed to stop instance {self.name}")

        logger.info(f"Successfully stopped instance {self.name}")

        # Remove stale host key(s) from known hosts
        self._filter_known_hosts()

        if delete_configs:
            # Delete from RNS
            rns_client.delete_configs(resource=self)
            logger.info(f"Deleted {self.name} from configs")

    def _sync_runhouse_to_cluster(self, node: str = None, _install_url=None, env=None):
        if not self.instance_id:
            raise ValueError(f"No instance ID set for cluster {self.name}. Is it up?")

        if not self._http_client:
            self.connect_server_client()

        # Sync the local ~/.rh directory to the cluster
        self._rsync(
            source=str(Path("~/.rh").expanduser()),
            dest="~/.rh",
            up=True,
            contents=True,
        )
        logger.info("Synced ~/.rh folder to the cluster")

        local_rh_package_path = Path(importlib.util.find_spec("runhouse").origin).parent
        # local_rh_package_path = Path(pkgutil.get_loader("runhouse").path).parent

        # **Note** temp patch to handle PyYAML errors: https://github.com/yaml/pyyaml/issues/724
        base_rh_install_cmd = f'{self._get_env_activate_cmd(env=None)} && python3 -m pip install "cython<3.0.0"'

        # Check if runhouse is installed from source and has setup.py
        if (
            not _install_url
            and local_rh_package_path.parent.name == "runhouse"
            and (local_rh_package_path.parent / "setup.py").exists()
        ):
            # Package is installed in editable mode
            local_rh_package_path = local_rh_package_path.parent
            dest_path = f"~/{local_rh_package_path.name}"

            self._rsync(
                source=str(local_rh_package_path),
                dest=dest_path,
                up=True,
                contents=True,
            )

            rh_install_cmd = (
                f"{base_rh_install_cmd} && python3 -m pip install ./runhouse[sagemaker]"
            )
        else:
            if not _install_url:
                import runhouse

                _install_url = f"runhouse[sagemaker]=={runhouse.__version__}"
            rh_install_cmd = (
                f"{base_rh_install_cmd} && python3 -m pip install {_install_url}"
            )

        status_codes = self.run([rh_install_cmd])

        if status_codes[0][0] != 0:
            raise ValueError(
                f"Error installing runhouse on cluster: {status_codes[0][1]}"
            )

    def _load_estimator(
        self, estimator: Union[Dict, "EstimatorBase", None]
    ) -> Union[None, "EstimatorBase"]:
        """Build an Estimator object from config"""
        if estimator is None:
            return None

        if isinstance(estimator, EstimatorBase):
            return estimator

        if not isinstance(estimator, dict):
            raise TypeError(
                f"Unsupported estimator type. Expected dictionary or EstimatorBase, got {type(estimator)}"
            )
        if "sagemaker_session" not in estimator:
            # Estimator requires an initialized sagemaker session
            # https://stackoverflow.com/questions/55869651/how-to-fix-aws-region-error-valueerror-must-setup-local-aws-configuration-with
            estimator["sagemaker_session"] = self._sagemaker_session

        # Re-build the estimator object from its config
        estimator_framework = self._estimator_framework
        if estimator_framework == "PyTorch":
            return PyTorch(**estimator)
        elif estimator_framework == "TensorFlow":
            return TensorFlow(**estimator)
        elif estimator_framework == "MXNet":
            return MXNet(**estimator)
        elif estimator_framework == "XGBoost":
            return XGBoost(**estimator)
        else:
            raise NotImplementedError(
                f"Unsupported estimator framework {estimator_framework}"
            )

    def _sync_estimator_to_cluster(self):
        """If providing a custom estimator sync over the estimator's source directory to the cluster"""
        from runhouse import folder

        estimator_folder = folder(
            path=Path(self._estimator_source_dir).expanduser()
        ).to(self, path=self.ESTIMATOR_SRC_CODE_PATH)
        logger.info(
            f"Synced estimator source directory to the cluster in path: {estimator_folder.path}"
        )

    def _base_image_uri(self):
        """Pick a default image for the cluster based on its instance type"""
        # TODO [JL] Add flexibility for py version & framework_version
        gpu_instance_types = ["p2", "p3", "p4", "g3", "g4", "g5"]

        image_type = (
            "gpu"
            if any(prefix in self.instance_type for prefix in gpu_instance_types)
            else "cpu"
        )

        cuda_version = "cu118-" if image_type == "gpu" else ""
        image_url = (
            f"{self.BASE_ECR_URL}/pytorch-training:2.0.1-{image_type}-py310-{cuda_version}"
            f"ubuntu20.04-sagemaker"
        )

        return image_url

    def _update_autostop(self, autostop_mins: int = None):
        cluster_config = self.config()
        cluster_config["autostop_mins"] = autostop_mins or -1
        if not self._http_client:
            self.connect_server_client()
        # Update the config on the server with the new autostop time
        self.client.check_server()

    # -------------------------------------------------------
    # Port Management
    # -------------------------------------------------------
    def _ports_are_in_use(self) -> bool:
        """Check if the ports used for port forwarding from localhost to the cluster are in use."""
        try:
            self._bind_ports_to_localhost()
            # Ports are not in use
            return False
        except OSError:
            # At least one of the ports is in use
            return True

    def _bind_ports_to_localhost(self):
        """Try binding the SSH and HTTP ports to localhost to check if they are in use."""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s1:
            s1.bind(("localhost", self.ssh_port))
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s2:
            s2.bind(("localhost", self.server_port))

    # -------------------------------------------------------
    # SSH config
    # -------------------------------------------------------
    def _filter_known_hosts(self):
        """To prevent host key collisions in the ~/.ssh/known_hosts file, remove any stale entries of localhost
        using the SSH port."""
        known_hosts = self.hosts_path
        if not known_hosts.exists():
            # e.g. in a collab or notebook environment
            return

        valid_hosts = []
        with open(known_hosts, "r") as f:
            for line in f:
                if not line.strip().startswith(
                    f"[{self.DEFAULT_SERVER_HOST}]:{self.ssh_port}"
                ):
                    valid_hosts.append(line)

        with open(known_hosts, "w") as f:
            f.writelines(valid_hosts)

    def _ssh_config_entry(
        self,
        port: int,
        name: str = None,
        hostname: str = None,
        identity_file: str = None,
        user: str = None,
    ):
        return textwrap.dedent(
            f"""
            # Added by Runhouse for SageMaker SSH Support
            Host {name or self.name}
              HostName {hostname or self.DEFAULT_SERVER_HOST}
              IdentityFile {identity_file or self._abs_ssh_key_path}
              Port {port}
              User {user or self.DEFAULT_USER}
        """
        )

    def _add_or_update_ssh_config_entry(self):
        """Update the SSH config to allow for accessing the cluster via: ssh <cluster name>"""
        connected_ssh_port = self.ssh_port
        config_file = self.ssh_config_file

        with open(config_file, "r") as f:
            existing_config = f.read()

        pattern = re.compile(
            rf"^\s*Host\s+{re.escape(self.name)}\s*$.*?(?=^\s*Host\s+|\Z)",
            re.MULTILINE | re.DOTALL,
        )

        entry_match = pattern.search(existing_config)

        if entry_match:
            # If entry already exists update the port with the connected SSH port (may have changed from previous
            # connection attempt)
            existing_entry = entry_match.group()
            updated_entry = re.sub(
                r"(?<=Port )\d+", str(connected_ssh_port), existing_entry
            )
            updated_config = existing_config.replace(existing_entry, updated_entry)
            with open(config_file, "w") as f:
                f.write(updated_config)
        else:
            # Otherwise, add the new entry to the config file
            new_entry = self._ssh_config_entry(port=connected_ssh_port)
            with open(config_file, "a") as f:
                f.write(new_entry)



import tempfile
from pathlib import Path

import pytest

import runhouse as rh
from runhouse.servers.http.http_utils import deserialize_data, serialize_data
from runhouse.servers.obj_store import ObjStore


@pytest.mark.servertest
class TestServlet:
    @pytest.mark.level("unit")
    def test_put_resource(self, test_servlet, blob_data):
        with tempfile.TemporaryDirectory() as temp_dir:
            resource_path = Path(temp_dir, "local-blob")
            local_blob = rh.blob(blob_data, path=resource_path)
            resource = local_blob.to(system="file", path=resource_path)

            state = {}
            resp = ObjStore.call_actor_method(
                test_servlet,
                "aput_resource_local",
                data=serialize_data(
                    (resource.config(condensed=False), state, resource.dryrun), "pickle"
                ),
                serialization="pickle",
            )

            assert resp.output_type == "result_serialized"
            assert deserialize_data(resp.data, resp.serialization).startswith("file_")

    @pytest.mark.level("unit")
    def test_put_obj_local(self, test_servlet, blob_data):
        with tempfile.TemporaryDirectory() as temp_dir:
            resource_path = Path(temp_dir, "local-blob")
            resource = rh.blob(blob_data, path=resource_path)
            resp = ObjStore.call_actor_method(
                test_servlet,
                "aput_local",
                key="key1",
                data=serialize_data(resource, "pickle"),
                serialization="pickle",
            )
            assert resp.output_type == "success"

    @pytest.mark.level("unit")
    def test_get_obj(self, test_servlet):
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aget_local",
            key="key1",
            default=KeyError,
            serialization="pickle",
            remote=False,
        )
        assert resp.output_type == "result_serialized"
        blob = deserialize_data(resp.data, resp.serialization)
        assert isinstance(blob, rh.Blob)

    @pytest.mark.level("unit")
    def test_get_obj_remote(self, test_servlet):
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aget_local",
            key="key1",
            default=KeyError,
            serialization="pickle",
            remote=True,
        )
        assert resp.output_type == "config"
        blob_config = deserialize_data(resp.data, resp.serialization)
        assert isinstance(blob_config, dict)

    @pytest.mark.level("unit")
    def test_get_obj_does_not_exist(self, test_servlet):
        resp = ObjStore.call_actor_method(
            test_servlet,
            "aget_local",
            key="abcdefg",
            default=KeyError,
            serialization="pickle",
            remote=False,
        )
        assert resp.output_type == "exception"
        error = deserialize_data(resp.data["error"], "pickle")
        assert isinstance(error, KeyError)



import argparse
import os
import shutil
import subprocess
import warnings

import dotenv

import requests

from git import Repo

dotenv.load_dotenv()

GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")

SOURCE_REPO_NAME = "runhouse"
SOURCE_REPO_PATH = f"run-house/{SOURCE_REPO_NAME}"
SOURCE_REPO_URL = f"https://github.com/{SOURCE_REPO_PATH}.git"

TARGET_REPO_NAME = "runhouse-docs"
TARGET_REPO_PATH = f"run-house/{TARGET_REPO_NAME}"

PATH_TO_DOCS = "runhouse/docs/_build"

HEADERS = {
    "Authorization": f"Bearer {GITHUB_TOKEN}",
    "Accept": "application/vnd.github.v3+json",
}


def clone_repo():
    # Clone the runhouse repo locally
    if os.path.exists(SOURCE_REPO_NAME):
        Repo(SOURCE_REPO_NAME)
    else:
        Repo.clone_from(SOURCE_REPO_URL, SOURCE_REPO_NAME)


def get_refs_from_repo(url):
    response = requests.get(url, headers=HEADERS)
    refs = response.json()
    return refs


def run_command(command):
    status_codes = subprocess.run(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    return status_codes


def build_and_copy_docs(branch_name, commit_hash=None):
    """Build docs based on the branch's latest commit (or tag revision) in the runhouse repo, and copy the
    resulting JSON files to the runhouse-docs repo."""
    try:
        if commit_hash:
            # Working with a particular release tag
            run_command(f"cd {SOURCE_REPO_NAME} && git reset --hard {commit_hash}")
        else:
            # Make sure the local runhouse folder is up to date with the latest remote branch
            run_command(
                f"cd {SOURCE_REPO_NAME} && git reset --hard origin/{branch_name}"
            )

        # Run the make json command in the "docs" directory of the local Runhouse repo
        res = run_command("cd runhouse && cd docs && make json")
        if res.returncode != 0:
            warnings.warn(
                f"Failed to build docs for branch: {branch_name}: {res.stderr}"
            )

        # Get the absolute paths for the source and destination folders
        abs_folder_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), PATH_TO_DOCS)
        )

        # Copy the contents of the generated JSON docs to the runhouse-docs local folder
        shutil.copytree(
            f"{abs_folder_path}/json",
            os.path.abspath(os.path.join(os.path.dirname(__file__), TARGET_REPO_NAME)),
        )

        run_command(
            f"cd {TARGET_REPO_NAME} && git init && git checkout -b {branch_name} "
            f"&& git remote add origin https://github.com/run-house/{TARGET_REPO_NAME}.git "
            f"&& git add . && git commit -m 'Updated docs from script'"
        )

        # Push changes to the remote runhouse-docs repo
        # Overwrite with whatever we have in the remote based on the latest version of the branch in the runhouse repo
        run_command(f"cd {TARGET_REPO_NAME} && git push --force origin {branch_name}")

        # Delete the _build folder in the local file system
        shutil.rmtree(abs_folder_path)

        # Delete the runhouse-docs folder in the local file system
        shutil.rmtree(
            os.path.abspath(os.path.join(os.path.dirname(__file__), TARGET_REPO_NAME))
        )

    except Exception as e:
        warnings.warn(f"Failed to build docs for {branch_name}: {str(e)}")


def generate_docs_for_branches():
    branches_url = f"https://api.github.com/repos/{SOURCE_REPO_PATH}/git/refs"
    refs = get_refs_from_repo(branches_url)

    # Build docs and copy JSON files for each branch
    for ref in refs:
        ref_type = ref["ref"].split("/")[-2]
        ref_name = ref["ref"].split("/")[-1]

        if ref_type == "heads":
            print(f"Building docs for branch: {ref_name}")
            # Check out locally to the given branch
            status_codes = run_command(f"cd runhouse && git checkout {ref_name}")
            if status_codes.returncode != 0:
                raise Exception(status_codes.stderr)

            build_and_copy_docs(ref_name)


def generate_docs_for_tags():
    # Handle tags (not releases)
    tags_url = f"https://api.github.com/repos/{SOURCE_REPO_PATH}/tags"
    releases = get_refs_from_repo(tags_url)
    for release in releases:
        tag_name = release["name"]
        tag_url = (
            f"https://api.github.com/repos/{SOURCE_REPO_PATH}/git/refs/tags/{tag_name}"
        )
        tag_response = requests.get(tag_url, headers=HEADERS)
        tag_info = tag_response.json()

        if "object" in tag_info and "sha" in tag_info["object"]:
            commit_hash = tag_info["object"]["sha"]
            print(f"Building docs for release: {tag_name} (commit hash: {commit_hash})")
            build_and_copy_docs(tag_name, commit_hash)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate docs for tags, branches, or both."
    )
    parser.add_argument(
        "--docs-type",
        nargs="?",
        default="tags",
        choices=["branches", "tags", "all"],
        help="Type of docs to build. If 'all' provided will build for both branches and tags. "
        "Default is 'tags'",
    )

    args = parser.parse_args()
    docs_type = args.docs_type

    clone_repo()

    if docs_type == "tags":
        generate_docs_for_tags()
    elif docs_type == "branches":
        generate_docs_for_branches()
    elif docs_type == "all":
        # run for branches & tags
        generate_docs_for_branches()
        generate_docs_for_tags()



from pathlib import Path

import pytest

import runhouse as rh

from runhouse import Cluster


TEMP_LOCAL_FOLDER = Path(__file__).parents[1] / "rh-blobs"


def test_save_local_blob_fails(local_blob, blob_data):
    with pytest.raises(ValueError):
        local_blob.save(name="my_local_blob")


@pytest.mark.parametrize(
    "blob",
    ["local_file", "s3_blob", "gcs_blob"],
    indirect=True,
)
def test_reload_blob_with_name(blob):
    name = "my_blob"
    blob.save(name)
    original_system = str(blob.system)
    original_data_str = str(blob.fetch())

    del blob

    reloaded_blob = rh.blob(name=name)
    assert str(reloaded_blob.system) == str(original_system)
    reloaded_data = reloaded_blob.fetch()
    assert reloaded_data[1] == "test"
    assert str(reloaded_data) == original_data_str

    # Delete metadata saved locally and / or the database for the blob
    reloaded_blob.delete_configs()

    # Delete the blob
    reloaded_blob.rm()
    assert not reloaded_blob.exists_in_system()


@pytest.mark.parametrize(
    "blob", ["local_file", "s3_blob", "gcs_blob", "cluster_file"], indirect=True
)
def test_reload_file_with_path(blob):
    reloaded_blob = rh.blob(path=blob.path, system=blob.system)
    reloaded_data = reloaded_blob.fetch()
    assert reloaded_data[1] == "test"

    # Delete the blob
    reloaded_blob.rm()
    assert not reloaded_blob.exists_in_system()


@pytest.mark.parametrize("file", ["local_file", "cluster_file"], indirect=True)
def test_file_to_blob(file, cluster):
    local_blob = file.to("here")
    assert local_blob.system is None
    fetched = local_blob.fetch()
    assert fetched[1] == "test"
    assert str(fetched) == str(file.fetch())

    cluster_blob = file.to(cluster)
    assert isinstance(cluster_blob.system, Cluster)
    fetched = cluster_blob.fetch()
    assert fetched[1] == "test"
    assert str(fetched) == str(file.fetch())


@pytest.mark.parametrize(
    "blob", ["local_blob", "cluster_blob", "local_file"], indirect=True
)
@pytest.mark.parametrize(
    "folder",
    ["local_folder", "cluster_folder", "s3_folder", "gcs_folder"],
    indirect=True,
)
def test_blob_to_file(blob, folder):
    new_file = blob.to(
        system=folder.system,
        path=folder.path + "/test_blob.pickle",
        data_config=folder.data_config,
    )
    assert new_file.system == folder.system
    assert new_file.path == folder.path + "/test_blob.pickle"
    fetched = new_file.fetch()
    assert fetched[1] == "test"
    assert str(fetched) == str(blob.fetch())
    assert "test_blob.pickle" in folder.ls(full_paths=False)


@pytest.mark.skip
def test_sharing_blob(cluster_blob):
    pass
    # TODO


def test_load_shared_blob(local_blob):
    my_blob = rh.blob(name="@/shared_blob")
    assert my_blob.exists_in_system()

    reloaded_data = my_blob.fetch()
    # NOTE: we need to do the deserialization ourselves
    assert str(reloaded_data) == str(local_blob.fetch())



"""Script used to keep the SageMaker cluster up, pending the autostop time provided in the cluster's config."""
import json
import os
import subprocess
import time
import warnings
from pathlib import Path
from typing import Dict

DEFAULT_AUTOSTOP = -1
MAIN_DIR = "/opt/ml/code"
OUT_FILE = "sm_cluster.out"

# ---------Configure SSH Helper----------
# https://github.com/aws-samples/sagemaker-ssh-helper#step-3-modify-your-training-script
import sagemaker_ssh_helper

sagemaker_ssh_helper.setup_and_start_ssh()


def run_training_job(path_to_job: str, num_attempts: int):
    job_succeeded = False

    try:
        # Execute the script as a separate process and capture stdout and stderr
        completed_process = subprocess.run(
            ["python", path_to_job],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,  # Redirect stderr to stdout
            text=True,
        )

        # Access combined stdout and stderr
        combined_output = completed_process.stdout

        if combined_output:
            print(combined_output)
            with open(f"{MAIN_DIR}/{OUT_FILE}", "w") as f:
                f.write(combined_output)

        job_succeeded = True

    except subprocess.CalledProcessError as e:
        warnings.warn(
            f"({e.returncode}) Error executing training script "
            f"(already made {num_attempts} attempts): {e.stderr}"
        )
        num_attempts += 1

    except Exception as e:
        num_attempts += 1
        warnings.warn(
            f"Error executing training script (already made {num_attempts} attempts): {e}"
        )

    finally:
        return job_succeeded, num_attempts


def read_cluster_config() -> Dict:
    """Read the autostop from the cluster's config - this will get populated when the cluster is created,
    or via the autostop APIs (e.g. `pause_autostop` or `keep_warm`)"""
    try:
        # Note: Runhouse has not yet been installed at this stage on the cluster,
        # so we can't import CLUSTER_CONFIG_PATH, we just need to hardcode it.
        with open(os.path.expanduser("~/.rh/cluster_config.json"), "r") as f:
            cluster_config = json.load(f)
    except FileNotFoundError:
        cluster_config = {}

    return cluster_config


if __name__ == "__main__":
    print("Launching instance from script")
    server_log = Path("~/.rh/server.log").expanduser()
    server_log.parent.mkdir(parents=True, exist_ok=True)
    server_log.touch()

    last_active = time.time()
    last_autostop_value = None
    training_job_completed = False
    path_to_job = None
    num_attempts = 0

    while True:
        last_active = server_log.stat().st_mtime
        config = read_cluster_config()

        autostop = int(config.get("autostop_mins", DEFAULT_AUTOSTOP))

        if autostop != -1:
            time_to_autostop = last_active + (autostop * 60)
            current_time = time.time()

            if current_time >= time_to_autostop:
                print("Autostop time reached, stopping instance")
                break

            # Reset launch time if autostop was updated
            if last_autostop_value is not None and autostop != last_autostop_value:
                print(f"Resetting autostop from {last_autostop_value} to {autostop}")
                last_active = current_time

            last_autostop_value = autostop

        estimator_entry_point = config.get("estimator_entry_point")
        if estimator_entry_point:
            # Update the path to the custom estimator job which was rsynced to the cluster
            # and now lives in path: /opt/ml/code
            path_to_job = f"{MAIN_DIR}/{estimator_entry_point}"

        if not training_job_completed and path_to_job:
            print(f"Running training job specified in path: {path_to_job}")
            training_job_completed, num_attempts = run_training_job(
                path_to_job, num_attempts
            )

        time.sleep(20)



from pathlib import Path
from typing import Dict, Optional, Union

import fsspec

from runhouse.logger import logger

from runhouse.resources.folders.folder import Folder
from runhouse.resources.hardware.utils import _get_cluster_from
from runhouse.resources.resource import Resource


def folder(
    name: Optional[str] = None,
    path: Optional[Union[str, Path]] = None,
    system: Optional[Union[str, "Cluster"]] = None,
    dryrun: bool = False,
    local_mount: bool = False,
    data_config: Optional[Dict] = None,
) -> Folder:
    """Creates a Runhouse folder object, which can be used to interact with the folder at the given path.

    Args:
        name (Optional[str]): Name to give the folder, to be re-used later on.
        path (Optional[str or Path]): Path (or path) that the folder is located at.
        system (Optional[str or Cluster]): File system or cluster name. If providing a file system this must be one of:
            [``file``, ``github``, ``sftp``, ``ssh``, ``s3``, ``gs``, ``azure``].
            We are working to add additional file system support.
        dryrun (bool): Whether to create the Folder if it doesn't exist, or load a Folder object as a dryrun.
            (Default: ``False``)
        local_mount (bool): Whether or not to mount the folder locally. (Default: ``False``)
        data_config (Optional[Dict]): The data config to pass to the underlying fsspec handler.

    Returns:
        Folder: The resulting folder.

    Example:
        >>> import runhouse as rh
        >>> rh.folder(name='training_imgs', path='remote_directory/images', system='s3').save()

        >>> # Load folder from above
        >>> reloaded_folder = rh.folder(name="training_imgs")
    """
    # TODO [DG] Include loud warning that relative paths are relative to the git root / working directory!

    if name and not any([path, system, local_mount, data_config]):
        # If only the name is provided
        try:
            return Folder.from_name(name, dryrun)
        except ValueError:
            # This is a rare instance where passing no constructor params is actually valid (anonymous folder),
            # so if we don't find the name, the user may still actually want to create a new folder.
            pass

    system = system or Folder.DEFAULT_FS
    if system == "s3":
        from .s3_folder import S3Folder

        return S3Folder(
            system=system,
            path=path,
            data_config=data_config,
            local_mount=local_mount,
            name=name,
            dryrun=dryrun,
        )
    elif system == "gs":
        from .gcs_folder import GCSFolder

        return GCSFolder(
            system=system,
            path=path,
            data_config=data_config,
            local_mount=local_mount,
            name=name,
            dryrun=dryrun,
        )
    elif system == "azure":
        from .azure_folder import AzureFolder

        return AzureFolder(
            system=system,
            path=path,
            data_config=data_config,
            local_mount=local_mount,
            name=name,
            dryrun=dryrun,
        )

    if system in fsspec.available_protocols():
        if system not in [Folder.DEFAULT_FS, "github", "sftp", "ssh"]:
            logger.warning(
                f"fsspec file system {system} not officially supported. Use at your own risk."
            )
        return Folder(
            system=system,
            path=path,
            data_config=data_config,
            local_mount=local_mount,
            name=name,
            dryrun=dryrun,
        )

    cluster_system = _get_cluster_from(system, dryrun=dryrun)
    if isinstance(cluster_system, Resource):
        return Folder(
            system=cluster_system,
            path=path,
            data_config=data_config,
            local_mount=local_mount,
            name=name,
            dryrun=dryrun,
        )

    raise ValueError(
        f"File system {system} not found. Have you installed the "
        f"necessary packages for this fsspec protocol? (e.g. s3fs for s3). If the file system "
        f"is a cluster (ex: /my-user/rh-cpu), make sure the cluster config has been saved."
    )



import numpy as np
import pytest

import runhouse as rh

from tests.conftest import init_args


@pytest.fixture
def blob(request):
    """Parametrize over multiple blobs - useful for running the same test on multiple storage types."""
    return request.getfixturevalue(request.param)


@pytest.fixture
def file(request):
    """Parametrize over multiple files - useful for running the same test on multiple storage types."""
    return request.getfixturevalue(request.param)


@pytest.fixture(scope="session")
def blob_data():
    return [np.arange(50), "test", {"a": 1, "b": 2}]


@pytest.fixture
def local_file(blob_data, tmp_path):
    args = {
        "data": blob_data,
        "system": "file",
        "path": str(tmp_path / "test_blob.pickle"),
    }
    b = rh.blob(**args)
    init_args[id(b)] = args
    return b


@pytest.fixture
def local_blob(blob_data):
    return rh.blob(
        data=blob_data,
    )


@pytest.fixture
def s3_blob(blob_data, blob_s3_bucket):
    return rh.blob(
        data=blob_data,
        system="s3",
        path=f"/{blob_s3_bucket}/test_blob.pickle",
    )


@pytest.fixture
def gcs_blob(blob_data, blob_gcs_bucket):
    return rh.blob(
        data=blob_data,
        system="gs",
        path=f"/{blob_gcs_bucket}/test_blob.pickle",
    )


@pytest.fixture
def cluster_blob(blob_data, ondemand_aws_cluster):
    return rh.blob(
        data=blob_data,
        system=ondemand_aws_cluster,
    )


@pytest.fixture
def cluster_file(blob_data, ondemand_aws_cluster):
    return rh.blob(
        data=blob_data,
        system=ondemand_aws_cluster,
        path="test_blob.pickle",
    )



import copy
import re
import sys
from pathlib import Path
from typing import Dict, Optional, Union

from runhouse.resources.envs.utils import install_conda, run_setup_command
from runhouse.resources.folders import Folder, folder
from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.hardware.utils import (
    _get_cluster_from,
    detect_cuda_version_or_cpu,
)
from runhouse.resources.resource import Resource
from runhouse.utils import (
    find_locally_installed_version,
    get_local_install_path,
    is_python_package_string,
    locate_working_dir,
)


INSTALL_METHODS = {"local", "reqs", "pip", "conda", "rh"}

from runhouse.logger import logger


class CodeSyncError(Exception):
    pass


class Package(Resource):
    RESOURCE_TYPE = "package"

    # https://pytorch.org/get-started/locally/
    # Note: no binaries exist for 11.4 (https://github.com/pytorch/pytorch/issues/75992)
    TORCH_INDEX_URLS = {
        "11.3": "https://download.pytorch.org/whl/cu113",
        "11.5": "https://download.pytorch.org/whl/cu115",
        "11.6": "https://download.pytorch.org/whl/cu116",
        "11.7": "https://download.pytorch.org/whl/cu117",
        "11.8": "https://download.pytorch.org/whl/cu118",
        "12.1": "https://download.pytorch.org/whl/cu121",
        "cpu": "https://download.pytorch.org/whl/cpu",
    }

    def __init__(
        self,
        name: str = None,
        install_method: str = None,
        install_target: Union[str, Folder] = None,
        install_args: str = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Package resource.

        .. note::
            To create a git package, please use the factory method :func:`package`.
        """
        super().__init__(
            name=name,
            dryrun=dryrun,
        )
        self.install_method = install_method
        self.install_target = install_target
        self.install_args = install_args

    def config(self, condensed=True):
        # If the package is just a simple Package.from_string string, no
        # need to store it in rns, just give back the string.
        # if self.install_method in ['pip', 'conda', 'git']:
        #     return f'{self.install_method}:{self.name}'
        config = super().config(condensed)
        config["install_method"] = self.install_method
        config["install_target"] = self._resource_string_for_subconfig(
            self.install_target, condensed
        )
        config["install_args"] = self.install_args
        return config

    def __str__(self):
        if self.name:
            return f"Package: {self.name}"
        if isinstance(self.install_target, Folder):
            # if self.install_target.name:
            #     return f'Package: {self.install_target.name}'
            return f"Package: {self.install_target.path}"
        return f"Package: {self.install_target}"

    @staticmethod
    def _prepend_python_executable(
        install_cmd: str, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        return (
            f"python3 -m {install_cmd}"
            if cluster or env
            else f"{sys.executable} -m {install_cmd}"
        )

    @staticmethod
    def _prepend_env_command(install_cmd: str, env: Union[str, "Env"] = None):
        if env:
            from runhouse.resources.envs.utils import _get_env_from

            env = _get_env_from(env)
            install_cmd = env._full_command(install_cmd)

        return install_cmd

    def _validate_folder_path(self):
        # If self.path is the same as the user's home directory, raise an error.
        # Check this with Path and expanduser to handle both relative and absolute paths.
        if Path(self.install_target.path).expanduser() in [
            Path("~").expanduser(),
            Path("/"),
        ]:
            raise CodeSyncError(
                "Cannot sync the home directory. Please include a Python configuration file in a subdirectory."
            )

    def _pip_install_cmd(
        self, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if isinstance(self.install_target, Folder):
            install_cmd = (
                f"{str(Path(self.install_target.local_path).absolute())}" + install_args
            )
        else:
            install_target = f'"{self.install_target}"'
            install_cmd = install_target + install_args

        install_cmd = f"pip install {self._install_cmd_for_torch(install_cmd, cluster)}"
        install_cmd = self._prepend_python_executable(
            install_cmd, cluster=cluster, env=env
        )
        install_cmd = self._prepend_env_command(install_cmd, env=env)
        return install_cmd

    def _conda_install_cmd(
        self, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if isinstance(self.install_target, Folder):
            install_cmd = f"{self.install_target.local_path}" + install_args
        else:
            install_cmd = self.install_target + install_args

        install_cmd = f"conda install -y {install_cmd}"
        install_cmd = self._prepend_env_command(install_cmd, env=env)
        install_conda(cluster=cluster)
        return install_cmd

    def _reqs_install_cmd(
        self, env: Union[str, "Env"] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if not isinstance(self.install_target, Folder):
            install_cmd = self.install_target + install_args
        else:
            path = self.install_target.local_path

            # Read reqs_path and reqs from local if not cluster
            if not cluster:
                reqs_path = f"{path}/requirements.txt"
                if not Path(reqs_path).expanduser().exists():
                    return None

                with open(reqs_path) as f:
                    reqs_list = f.readlines()

            # Otherwise, make sure the target folder is on the cluster,
            # and read reqs from the cluster
            else:
                if not self.install_target.system == cluster:
                    install_target = self.to(cluster).install_target
                else:
                    install_target = self.install_target

                if not install_target.exists_in_system():
                    return None
                elif "requirements.txt" not in install_target.ls(full_paths=False):
                    return None

                reqs_list = (
                    install_target.get("requirements.txt", mode="r")
                    .strip("\n")
                    .split("\n")
                )
                reqs_path = f"{install_target.path}/requirements.txt"

            install_cmd = self._reqs_install_cmd_for_torch(
                reqs_path, reqs_list, install_args, cluster=cluster
            )

        install_cmd = f"pip install {install_cmd}"
        install_cmd = self._prepend_python_executable(
            install_cmd, env=env, cluster=cluster
        )
        install_cmd = self._prepend_env_command(install_cmd, env=env)
        return install_cmd

    def _install(self, env: Union[str, "Env"] = None, cluster: "Cluster" = None):
        """Install package.

        Args:
            env (Env or str): Environment to install package on. If left empty, defaults to base environment.
                (Default: ``None``)
            cluster (Optional[Cluster]): If provided, will install package on cluster using SSH.
        """

        logger.info(f"Installing {str(self)} with method {self.install_method}.")

        if isinstance(self.install_target, Folder):
            if not cluster:
                path = self.install_target.local_path
            elif self.install_target.exists_in_system():
                path = self.install_target.path
            else:
                path = self.to(cluster).install_target.path

            if not path:
                return

        if self.install_method == "pip":
            install_cmd = self._pip_install_cmd(env=env, cluster=cluster)
            logger.info(f"Running via install_method pip: {install_cmd}")
            retcode = run_setup_command(install_cmd, cluster=cluster)[0]
            if retcode != 0:
                raise RuntimeError(
                    f"Pip install {install_cmd} failed, check that the package exists and is available for your platform."
                )

        elif self.install_method == "conda":
            install_cmd = self._conda_install_cmd(env=env, cluster=cluster)
            logger.info(f"Running via install_method conda: {install_cmd}")
            retcode = run_setup_command(install_cmd, cluster=cluster)[0]
            if retcode != 0:
                raise RuntimeError(
                    f"Conda install {install_cmd} failed, check that the package exists and is "
                    "available for your platform."
                )

        elif self.install_method == "reqs":
            install_cmd = self._reqs_install_cmd(env=env, cluster=cluster)
            if install_cmd:
                logger.info(f"Running via install_method reqs: {install_cmd}")
                retcode = run_setup_command(install_cmd, cluster=cluster)[0]
                if retcode != 0:
                    raise RuntimeError(
                        f"Reqs install {install_cmd} failed, check that the package exists and is available for your platform."
                    )
            else:
                logger.info(f"{path}/requirements.txt not found, skipping reqs install")

        else:
            if self.install_method != "local":
                raise ValueError(
                    f"Unknown install method {self.install_method}. Must be one of {INSTALL_METHODS}"
                )

        # Need to append to path
        if self.install_method in ["local", "reqs"]:
            if isinstance(self.install_target, Folder):
                sys.path.insert(0, path) if not cluster else run_setup_command(
                    f"export PATH=$PATH;{path}", cluster=cluster
                )
            elif not cluster:
                if Path(self.install_target).resolve().expanduser().exists():
                    sys.path.insert(
                        0, str(Path(self.install_target).resolve().expanduser())
                    )
                else:
                    raise ValueError(
                        f"install_target {self.install_target} must be a Folder or a path to a directory for install_method {self.install_method}"
                    )
            else:
                raise ValueError(
                    f"If cluster is provided, install_target must be a Folder for install_method {self.install_method}"
                )

    # ----------------------------------
    # Torch Install Helpers
    # ----------------------------------
    def _reqs_install_cmd_for_torch(
        self, reqs_path, reqs_list, install_args="", cluster=None
    ):
        """Read requirements from file, append --index-url and --extra-index-url where relevant for torch packages,
        and return list of formatted packages."""
        # if torch extra index url is already defined by the user or torch isn't a req, directly pip install reqs file
        if not any("torch" in req for req in reqs_list):
            return f"-r {reqs_path}" + install_args

        cuda_version_or_cpu = detect_cuda_version_or_cpu(cluster=cluster)
        for req in reqs_list:
            if (
                "--index-url" in req or "--extra-index-url" in req
            ) and "pytorch.org" in req:
                return f"-r {reqs_path}" + install_args

        # add extra-index-url for torch if not found
        return f"-r {reqs_path} --extra-index-url {self._torch_index_url(cuda_version_or_cpu)}"

    def _install_cmd_for_torch(self, install_cmd, cluster=None):
        """Return the correct formatted pip install command for the torch package(s) provided."""
        if install_cmd.startswith("#"):
            return None

        torch_source_packages = ["torch", "torchvision", "torchaudio"]
        if not any([x in install_cmd for x in torch_source_packages]):
            return install_cmd

        packages_to_install: list = self._packages_to_install_from_cmd(install_cmd)
        final_install_cmd = ""
        cuda_version_or_cpu = detect_cuda_version_or_cpu(cluster=cluster)
        for package_install_cmd in packages_to_install:
            formatted_cmd = self._install_url_for_torch_package(
                package_install_cmd, cuda_version_or_cpu
            )
            if formatted_cmd:
                final_install_cmd += formatted_cmd + " "

        final_install_cmd = final_install_cmd.rstrip()
        return final_install_cmd if final_install_cmd != "" else None

    def _install_url_for_torch_package(self, install_cmd, cuda_version_or_cpu):
        """Build the full install command, adding a --index-url and --extra-index-url where applicable."""
        # Grab the relevant index url for torch based on the CUDA version provided
        if "," in install_cmd:
            # If installing a range of versions format the string to make it compatible with `pip_install` method
            install_cmd = install_cmd.replace(" ", "")

        index_url = self._torch_index_url(cuda_version_or_cpu)
        if index_url and not any(
            specifier in install_cmd for specifier in ["--index-url ", "-i "]
        ):
            install_cmd = f"{install_cmd} --index-url {index_url}"

        if "--extra-index-url" not in install_cmd:
            return f"{install_cmd} --extra-index-url https://pypi.python.org/simple/"

        return install_cmd

    def _torch_index_url(self, cuda_version_or_cpu: str):
        return self.TORCH_INDEX_URLS.get(cuda_version_or_cpu)

    @staticmethod
    def _packages_to_install_from_cmd(install_cmd: str):
        """Split a string of command(s) into a list of separate commands"""
        # Remove any --extra-index-url flags from the install command (to be added later by default)
        install_cmd = re.sub(r"--extra-index-url\s+\S+", "", install_cmd)
        install_cmd = install_cmd.strip()

        if ", " in install_cmd:
            # Ex: 'torch>=1.13.0,<2.0.0'
            return [install_cmd]

        matches = re.findall(r"(\S+(?:\s+(-i|--index-url)\s+\S+)?)", install_cmd)

        packages_to_install = [match[0] for match in matches]
        return packages_to_install

    def to(
        self,
        system: Union[str, Dict, "Cluster"],
        path: Optional[str] = None,
        mount: bool = False,
    ):
        """Copy the package onto filesystem or cluster, and return the new Package object."""
        if not isinstance(self.install_target, Folder):
            raise TypeError(
                "`install_target` must be a Folder in order to copy the package to a system."
            )

        if (
            isinstance(self.install_target.system, str)
            and not self.install_target.system == "file"
        ):
            self.install_target.system = _get_cluster_from(self.install_target.system)

        install_system_name = (
            self.install_target.system.name
            if isinstance(self.install_target.system, Cluster)
            else self.install_target.system
        )
        system = _get_cluster_from(system)
        system_name = system.name if isinstance(system, Cluster) else system

        if system_name == install_system_name:
            return self

        if isinstance(system, Resource):
            if isinstance(self.install_target.system, Resource):
                # if these are both clusters, check if they're pointing to the same address.
                # We use endpoint instead of address here because if address is localhost, we need to port too
                if self.install_target.system.endpoint(
                    external=False
                ) == system.endpoint(external=False):
                    # If we're on the target system, just make sure the package is in the Python path
                    sys.path.insert(0, self.install_target.local_path)
                    return self
            logger.info(
                f"Copying package from {self.install_target.fsspec_url} to: {getattr(system, 'name', system)}"
            )
            self._validate_folder_path()
            new_folder = self.install_target._to_cluster(system, path=path, mount=mount)
        else:  # to fs
            self._validate_folder_path()
            new_folder = self.install_target.to(system, path=path)
        new_folder.system = system
        new_package = copy.copy(self)
        new_package.install_target = new_folder
        return new_package

    @staticmethod
    def split_req_install_method(req_str: str):
        """Split a requirements string into a install method and the rest of the string."""
        splat = req_str.split(":", 1)
        return (splat[0], splat[1]) if len(splat) > 1 else ("", splat[0])

    @staticmethod
    def from_config(config: dict, dryrun=False, _resolve_children=True):
        if isinstance(config.get("install_target"), dict):
            config["install_target"] = Folder.from_config(
                config["install_target"],
                dryrun=dryrun,
                _resolve_children=_resolve_children,
            )
        if config.get("resource_subtype") == "GitPackage":
            from runhouse import GitPackage

            return GitPackage.from_config(
                config, dryrun=dryrun, _resolve_children=_resolve_children
            )
        return Package(**config, dryrun=dryrun)

    @staticmethod
    def from_string(specifier: str, dryrun=False):
        if specifier == "requirements.txt":
            specifier = "reqs:./"

        # Use regex to check if specifier matches '<method>:https://github.com/<path>' or 'https://github.com/<path>'
        match = re.search(
            r"^(?:(?P<method>[^:]+):)?(?P<path>https://github.com/.+)", specifier
        )
        if match:
            install_method = match.group("method")
            url = match.group("path")
            from runhouse.resources.packages.git_package import git_package

            return git_package(
                git_url=url, install_method=install_method, dryrun=dryrun
            )

        install_method, target_and_args = Package.split_req_install_method(specifier)

        # Handles a case like "torch --index-url https://download.pytorch.org/whl/cu113"
        rel_target, args = (
            target_and_args.split(" ", 1)
            if " " in target_and_args
            else (target_and_args, "")
        )

        # We need to do this because relative paths are relative to the current working directory!
        abs_target = (
            Path(rel_target).expanduser()
            if Path(rel_target).expanduser().is_absolute()
            else Path(locate_working_dir()) / rel_target
        )
        if abs_target.exists():
            target = Folder(
                path=abs_target, dryrun=True
            )  # No need to create the folder here
        else:
            target = rel_target

        # If install method is not provided, we need to infer it
        if not install_method:
            if Path(specifier).resolve().exists():
                install_method = "reqs"
            else:
                install_method = "pip"

        # If we are just defaulting to pip, attempt to install the same version of the package
        # that is already installed locally
        # Check if the target is only letters, nothing else. This means its a string like 'numpy'.
        if install_method == "pip" and is_python_package_string(target):
            locally_installed_version = find_locally_installed_version(target)
            if locally_installed_version:
                # Check if this is a package that was installed from local
                local_install_path = get_local_install_path(target)
                if local_install_path and Path(local_install_path).exists():
                    target = Folder(path=local_install_path, dryrun=True)

        # "Local" install method is a special case where we just copy a local folder and add to path
        if install_method == "local":
            return Package(
                install_target=target, install_method=install_method, dryrun=dryrun
            )

        elif install_method in ["reqs", "pip", "conda"]:
            return Package(
                install_target=target,
                install_args=args,
                install_method=install_method,
                dryrun=dryrun,
            )
        elif install_method == "rh":
            # Calling the factory method below
            return package(name=specifier[len("rh:") :], dryrun=dryrun)
        else:
            raise ValueError(
                f"Unknown install method {install_method}. Must be one of {INSTALL_METHODS}"
            )


def package(
    name: str = None,
    install_method: str = None,
    install_str: str = None,
    path: str = None,
    system: str = None,
    dryrun: bool = False,
    local_mount: bool = False,
    data_config: Optional[Dict] = None,
) -> Package:
    """
    Builds an instance of :class:`Package`.

    Args:
        name (str): Name to assign the package resource.
        install_method (str): Method for installing the package. Options: [``pip``, ``conda``, ``reqs``, ``local``]
        install_str (str): Additional arguments to install.
        path (str): URL of the package to install.
        system (str): File system or cluster on which the package lives. Currently this must a cluster or one of:
            [``file``, ``github``, ``sftp``, ``ssh``, ``s3``, ``gs``, ``azure``].
        dryrun (bool): Whether to create the Package if it doesn't exist, or load the Package object as a dryrun.
            (Default: ``False``)
        local_mount (bool): Whether to locally mount the installed package. (Default: ``False``)
        data_config (Optional[Dict]): The data config to pass to the underlying fsspec handler.

    Returns:
        Package: The resulting package.

    Example:
        >>> import runhouse as rh
        >>> reloaded_package = rh.package(name="my-package")
        >>> local_package = rh.package(path="local/folder/path", install_method="local")
    """
    if name and not any(
        [install_method, install_str, path, system, data_config, local_mount]
    ):
        # If only the name is provided and dryrun is set to True
        return Package.from_name(name, dryrun)

    install_target = None
    install_args = None
    if path is not None:
        system = system or Folder.DEFAULT_FS
        install_target = folder(
            path=path, system=system, local_mount=local_mount, data_config=data_config
        )
        install_args = install_str
    elif install_str is not None:
        install_target, install_args = install_str.split(" ", 1)

    return Package(
        install_method=install_method,
        install_target=install_target,
        install_args=install_args,
        name=name,
        dryrun=dryrun,
    )



from datetime import datetime
from pathlib import Path

from runhouse.globals import configs, rns_client
from runhouse.resources.hardware.utils import _get_cluster_from

DEFAULT_LOCAL_FOLDER = f"{Path.cwd()}/"
DEFAULT_CLUSTER_FS_FOLDER = (
    ""  # Objects will land inside home directory when sent without a path
)
DEFAULT_BLOB_STORAGE_FOLDER = (
    configs.get("default_blob_storage_folder", "runhouse") + "/"
)


def _generate_default_name(prefix: str = None, precision: str = "s", sep="_") -> str:
    """Name of the Run's parent folder which contains the Run's data (config, stdout, stderr, etc).
    If a name is provided, prepend that to the current timestamp to complete the folder name."""
    if precision == "d":
        timestamp_key = f"{datetime.now().strftime('%Y%m%d')}"
    elif precision == "s":
        timestamp_key = f"{datetime.now().strftime(f'%Y%m%d{sep}%H%M%S')}"
    elif precision == "ms":
        timestamp_key = f"{datetime.now().strftime(f'%Y%m%d{sep}%H%M%S_%f')}"
    if prefix is None:
        return timestamp_key
    return f"{prefix}{sep}{timestamp_key}"


def _generate_default_path(cls, name, system):
    """Generate a default path for a data resource. Logic is as follows:
    1. If the system is a local file system, save to the current working directory
    2. If the system is a remote file system, save to the default cache folder
    3. If the system is a remote object store, save to the default object store folder
    """

    from runhouse.resources.hardware import Cluster

    system = _get_cluster_from(system)

    name = name or _generate_default_name(prefix=cls.RESOURCE_TYPE)
    if system == rns_client.DEFAULT_FS or "here":
        base_folder = DEFAULT_LOCAL_FOLDER
    elif isinstance(system, Cluster):
        if system.on_this_cluster():
            base_folder = DEFAULT_LOCAL_FOLDER
        else:
            base_folder = DEFAULT_CLUSTER_FS_FOLDER
    else:
        base_folder = DEFAULT_BLOB_STORAGE_FOLDER
    return f"{base_folder}{cls.RESOURCE_TYPE}/{name}"



from .cluster import Cluster
from .cluster_factory import (
    cluster,
    kubernetes_cluster,
    ondemand_cluster,
    sagemaker_cluster,
)
from .on_demand_cluster import OnDemandCluster
from .sagemaker.sagemaker_cluster import SageMakerCluster
from .utils import (
    _current_cluster,
    _default_env_if_on_cluster,
    _get_cluster_from,
    cluster_config_file_exists,
    load_cluster_config_from_file,
)



import asyncio
import copy
import json
import threading
from typing import Any, Dict, List, Optional, Set, Union

import requests

import runhouse

from runhouse.constants import (
    DEFAULT_LOG_LEVEL,
    DEFAULT_LOG_SURFACING_INTERVAL,
    DEFAULT_STATUS_CHECK_INTERVAL,
    INCREASED_INTERVAL,
    INCREASED_STATUS_CHECK_INTERVAL,
    S3_LOGS_FILE_NAME,
    SCHEDULERS_DELAY,
    SERVER_LOGFILE,
)

from runhouse.globals import configs, obj_store, rns_client
from runhouse.logger import ColoredFormatter, logger
from runhouse.resources.hardware import load_cluster_config_from_file
from runhouse.resources.hardware.utils import detect_cuda_version_or_cpu
from runhouse.rns.rns_client import ResourceStatusData
from runhouse.rns.utils.api import ResourceAccess
from runhouse.servers.autostop_helper import AutostopHelper
from runhouse.servers.http.auth import AuthCache

from runhouse.utils import sync_function


class ClusterServletError(Exception):
    pass


class ClusterServlet:
    async def __init__(
        self, cluster_config: Optional[Dict[str, Any]] = None, *args, **kwargs
    ):
        # We do this here instead of at the start of the HTTP Server startup
        # because someone can be running `HTTPServer()` standalone in a test
        # and still want an initialized cluster config in the servlet.
        if not cluster_config:
            cluster_config = load_cluster_config_from_file()

        self.cluster_config: Optional[Dict[str, Any]] = (
            cluster_config if cluster_config else {}
        )
        self._initialized_env_servlet_names: Set[str] = set()
        self._key_to_env_servlet_name: Dict[Any, str] = {}
        self._auth_cache: AuthCache = AuthCache(cluster_config)
        self.autostop_helper = None

        logger.setLevel(kwargs.get("logs_level", DEFAULT_LOG_LEVEL))
        self.logger = logger

        if cluster_config.get("resource_subtype", None) == "OnDemandCluster":
            self.autostop_helper = AutostopHelper()

        logger.info("Creating periodic_cluster_checks thread.")
        cluster_checks_thread = threading.Thread(
            target=self.periodic_cluster_checks, daemon=True
        )
        cluster_checks_thread.start()

    ##############################################
    # Cluster config state storage methods
    ##############################################
    async def aget_cluster_config(self) -> Dict[str, Any]:
        return self.cluster_config

    async def aset_cluster_config(self, cluster_config: Dict[str, Any]):
        if "has_cuda" not in cluster_config.keys():
            cluster_config["has_cuda"] = detect_cuda_version_or_cpu() != "cpu"

        self.cluster_config = cluster_config

        # Propagate the changes to all other process's obj_stores
        await asyncio.gather(
            *[
                obj_store.acall_env_servlet_method(
                    env_servlet_name,
                    "aset_cluster_config",
                    cluster_config,
                    use_env_servlet_cache=False,
                )
                for env_servlet_name in await self.aget_all_initialized_env_servlet_names()
            ]
        )

        return self.cluster_config

    async def aset_cluster_config_value(self, key: str, value: Any):
        if self.autostop_helper and key == "autostop_mins":
            await self.autostop_helper.set_autostop(value)
        self.cluster_config[key] = value

        # Propagate the changes to all other process's obj_stores
        await asyncio.gather(
            *[
                obj_store.acall_env_servlet_method(
                    env_servlet_name,
                    "aset_cluster_config_value",
                    key,
                    value,
                    use_env_servlet_cache=False,
                )
                for env_servlet_name in await self.aget_all_initialized_env_servlet_names()
            ]
        )

        return self.cluster_config

    ##############################################
    # Auth cache internal functions
    ##############################################
    async def aresource_access_level(
        self, token: str, resource_uri: str
    ) -> Union[str, None]:
        # If the token in this request matches that of the owner of the cluster,
        # they have access to everything
        if configs.token and (
            configs.token == token
            or rns_client.cluster_token(configs.token, resource_uri) == token
        ):
            return ResourceAccess.WRITE
        return self._auth_cache.lookup_access_level(token, resource_uri)

    async def aget_username(self, token: str) -> str:
        return self._auth_cache.get_username(token)

    async def ahas_resource_access(self, token: str, resource_uri=None) -> bool:
        """Checks whether user has read or write access to a given module saved on the cluster."""
        from runhouse.rns.utils.api import ResourceAccess

        if token is None:
            # If no token is provided assume no access
            return False

        cluster_uri = self.cluster_config["name"]
        cluster_access = await self.aresource_access_level(token, cluster_uri)
        if cluster_access == ResourceAccess.WRITE:
            # if user has write access to cluster will have access to all resources
            return True

        if resource_uri is None and cluster_access not in [
            ResourceAccess.WRITE,
            ResourceAccess.READ,
        ]:
            # If module does not have a name, must have access to the cluster
            return False

        resource_access_level = await self.aresource_access_level(token, resource_uri)
        if resource_access_level not in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return False

        return True

    async def aclear_auth_cache(self, token: str = None):
        self._auth_cache.clear_cache(token)

    ##############################################
    # Key to servlet where it is stored mapping
    ##############################################
    async def amark_env_servlet_name_as_initialized(self, env_servlet_name: str):
        self._initialized_env_servlet_names.add(env_servlet_name)

    async def ais_env_servlet_name_initialized(self, env_servlet_name: str) -> bool:
        return env_servlet_name in self._initialized_env_servlet_names

    async def aget_all_initialized_env_servlet_names(self) -> Set[str]:
        return self._initialized_env_servlet_names

    async def aget_key_to_env_servlet_name_dict_keys(self) -> List[Any]:
        return list(self._key_to_env_servlet_name.keys())

    async def aget_key_to_env_servlet_name_dict(self) -> Dict[Any, str]:
        return self._key_to_env_servlet_name

    async def aget_env_servlet_name_for_key(self, key: Any) -> str:
        if self.autostop_helper:
            await self.autostop_helper.set_last_active_time_to_now()
        return self._key_to_env_servlet_name.get(key, None)

    async def aput_env_servlet_name_for_key(self, key: Any, env_servlet_name: str):
        if not await self.ais_env_servlet_name_initialized(env_servlet_name):
            raise ValueError(
                f"Env servlet name {env_servlet_name} not initialized, and you tried to mark a resource as in it."
            )
        self._key_to_env_servlet_name[key] = env_servlet_name

    async def apop_env_servlet_name_for_key(self, key: Any, *args) -> str:
        # *args allows us to pass default or not
        return self._key_to_env_servlet_name.pop(key, *args)

    async def aclear_key_to_env_servlet_name_dict(self):
        self._key_to_env_servlet_name = {}

    ##############################################
    # Remove Env Servlet
    ##############################################
    async def aclear_all_references_to_env_servlet_name(self, env_servlet_name: str):
        self._initialized_env_servlet_names.remove(env_servlet_name)
        deleted_keys = [
            key
            for key, env in self._key_to_env_servlet_name.items()
            if env == env_servlet_name
        ]
        for key in deleted_keys:
            self._key_to_env_servlet_name.pop(key)
        return deleted_keys

    ##############################################
    # Cluster status functions
    ##############################################

    async def aperiodic_cluster_checks(self):
        # Delay the start of post_status_thread, so we'll finish the cluster startup properly
        await asyncio.sleep(SCHEDULERS_DELAY)
        while True:
            try:

                cluster_config = await self.aget_cluster_config()
                interval_size = cluster_config.get(
                    "status_check_interval", DEFAULT_STATUS_CHECK_INTERVAL
                )

                # Only if one of these is true, do we actually need to get the status from each EnvServlet
                should_send_status_and_logs_to_den = (
                    configs.token is not None and interval_size != -1
                )
                should_update_autostop = self.autostop_helper is not None

                # turning off sending logs for now, until latency issue resolved and send logs revamp will land in next release
                send_logs = False

                if should_send_status_and_logs_to_den or should_update_autostop:
                    logger.info(
                        "Performing cluster checks: potentially sending to Den, surfacing logs to Den or updating autostop."
                    )
                    status: ResourceStatusData = await self.astatus()

                    if should_update_autostop:
                        function_running = any(
                            any(
                                len(
                                    resource["env_resource_mapping"][resource_name].get(
                                        "active_function_calls", []
                                    )
                                )
                                > 0
                                for resource_name in resource[
                                    "env_resource_mapping"
                                ].keys()
                            )
                            for resource in status.env_servlet_processes.values()
                        )
                        if function_running:
                            await self.autostop_helper.set_last_active_time_to_now()
                        # We do this separately from the set_last_active_time_to_now call above because
                        # function_running will only reflect activity from functions which happen to be running during
                        # the status check. We still need to attempt to register activity for functions which have
                        # been called and completed.
                        await self.autostop_helper.register_activity_if_needed()

                    if should_send_status_and_logs_to_den:
                        cluster_rns_address = cluster_config.get("name")
                        cluster_uri = rns_client.format_rns_address(cluster_rns_address)
                        api_server_url = status.cluster_config.get(
                            "api_server_url", rns_client.api_server_url
                        )

                        sent_status = await rns_client.send_status(
                            status=status,
                            cluster_uri=cluster_uri,
                            api_server_url=api_server_url,
                        )

                        if sent_status == 404:
                            logger.info(
                                "Cluster has not been saved, no update to status or logs."
                            )
                        elif sent_status != 200:
                            logger.error(
                                f"{sent_status}: Error in sending cluster status to Den. Check cluster logs for more info."
                            )
                        else:
                            logger.debug("Successfully sent cluster status to Den.")

                            if send_logs:
                                logs_resp = await rns_client.send_cluster_logs_to_den(
                                    cluster_uri=cluster_uri,
                                    api_server_url=api_server_url,
                                )

                                if logs_resp != 200:
                                    logger.error(
                                        f"{logs_resp}: Error in sending cluster logs to Den. Check cluster logs for more info."
                                    )
                                else:
                                    logger.debug(
                                        "Successfully sent cluster logs to Den."
                                    )

            except Exception as e:
                self.logger.error(
                    f"Cluster status check has failed: {str(e)}. Please check cluster logs for more info."
                )
                self.logger.warning(
                    f"Temporarily increasing the interval between two consecutive status checks. "
                    f"Next status check will be in {round(INCREASED_INTERVAL / 60, 2)} minutes. "
                    f"For changing the interval size, please run cluster._enable_or_update_status_check(new_interval). "
                    f"If a value is not provided, interval size will be set to {round(DEFAULT_STATUS_CHECK_INTERVAL / 60, 2)} minutes."
                )
                await asyncio.sleep(INCREASED_STATUS_CHECK_INTERVAL)
            else:
                # make sure that the thread will go to sleep, even if the interval size == -1
                # (meaning that sending status to den is disabled).
                interval_size = (
                    DEFAULT_STATUS_CHECK_INTERVAL
                    if interval_size == -1
                    else interval_size
                )
                await asyncio.sleep(interval_size)

    def periodic_cluster_checks(self):
        # This is only ever called once in its own thread, so we can do asyncio.run here instead of
        # sync_function.
        asyncio.run(self.aperiodic_cluster_checks())

    async def _status_for_env_servlet(self, env_servlet_name):
        try:
            (
                objects_in_env_servlet,
                env_servlet_utilization_data,
            ) = await obj_store.acall_env_servlet_method(
                env_servlet_name, method="astatus_local"
            )

            return {
                "env_servlet_name": env_servlet_name,
                "objects_in_env_servlet": objects_in_env_servlet,
                "env_servlet_utilization_data": env_servlet_utilization_data,
            }

        # Need to catch the exception here because we're running this in a gather,
        # and need to know which env servlet failed
        except Exception as e:
            return {"env_servlet_name": env_servlet_name, "Exception": e}

    async def astatus(self):
        import psutil

        from runhouse.utils import get_pid

        config_cluster = copy.deepcopy(self.cluster_config)

        # Popping out creds because we don't want to show them in the status
        config_cluster.pop("creds", None)

        # Getting data from each env servlet about the objects it contains and the utilization data
        env_servlet_utilization_data = {}
        env_servlets_status = await asyncio.gather(
            *[
                self._status_for_env_servlet(env_servlet_name)
                for env_servlet_name in self._initialized_env_servlet_names
            ],
        )

        # Store the data for the appropriate env servlet name
        for env_status in env_servlets_status:
            env_servlet_name = env_status.get("env_servlet_name")

            # Nothing if there was an exception
            if "Exception" in env_status.keys():
                e = env_status.get("Exception")
                self.logger.warning(
                    f"Exception {str(e)} in status for env servlet {env_servlet_name}"
                )
                env_servlet_utilization_data[env_servlet_name] = {}

            # Otherwise, store what was in the env and the utilization data
            else:
                env_memory_info = env_status.get("env_servlet_utilization_data")
                env_memory_info["env_resource_mapping"] = env_status.get(
                    "objects_in_env_servlet"
                )
                env_servlet_utilization_data[env_servlet_name] = env_memory_info

        # TODO: decide if we need this info at all: cpu_usage, memory_usage, disk_usage
        cpu_usage = psutil.cpu_percent(interval=1)

        # Fields: `available`, `percent`, `used`, `free`, `active`, `inactive`, `buffers`, `cached`, `shared`, `slab`
        memory_usage = psutil.virtual_memory()._asdict()

        # Fields: `total`, `used`, `free`, `percent`
        disk_usage = psutil.disk_usage("/")._asdict()

        status_data = {
            "cluster_config": config_cluster,
            "runhouse_version": runhouse.__version__,
            "server_pid": get_pid(),
            "env_servlet_processes": env_servlet_utilization_data,
            "system_cpu_usage": cpu_usage,
            "system_memory_usage": memory_usage,
            "system_disk_usage": disk_usage,
        }
        status_data = ResourceStatusData(**status_data)
        return status_data

    def status(self):
        return sync_function(self.astatus)()

    ##############################################
    # Surface cluster logs to Den
    ##############################################
    def _get_logs(self):
        with open(SERVER_LOGFILE) as log_file:
            log_lines = log_file.readlines()
        cleaned_log_lines = [ColoredFormatter.format_log(line) for line in log_lines]
        return " ".join(cleaned_log_lines)

    async def asend_cluster_logs_to_den(self):
        # Delay the start of post_logs_thread, so we'll finish the cluster startup properly
        await asyncio.sleep(SCHEDULERS_DELAY)

        while True:
            self.logger.info("Trying to send cluster logs to Den")
            try:
                interval_size = DEFAULT_LOG_SURFACING_INTERVAL
                latest_logs = self._get_logs()
                logs_data = {"file_name": S3_LOGS_FILE_NAME, "logs": latest_logs}

                cluster_config = await self.aget_cluster_config()
                cluster_uri = rns_client.format_rns_address(cluster_config.get("name"))
                api_server_url = cluster_config.get(
                    "api_server_url", rns_client.api_server_url
                )

                post_logs_resp = requests.post(
                    f"{api_server_url}/resource/{cluster_uri}/logs",
                    data=json.dumps(logs_data),
                    headers=rns_client.request_headers(),
                )

                post_logs_resp_json = post_logs_resp.json()

                if post_logs_resp.status_code != 200:
                    post_logs_error = (
                        post_logs_resp_json.get("detail")
                        if post_logs_resp_json.get("detail")
                        else ""
                    )
                    self.logger.error(
                        f"({post_logs_resp.status_code}) Failed to send cluster logs to Den: {post_logs_error}"
                    )
                else:
                    self.logger.debug(
                        f"Successfully sent cluster logs to Den. Next status check will be in {round(interval_size / 60, 2)} minutes."
                    )
            except Exception as e:
                self.logger.error(
                    f"Sending cluster logs to den has failed: {e}. Please check cluster logs for more info."
                )
                self.logger.warning(
                    f"Temporarily increasing the interval between two consecutive log retrievals."
                    f"Next log retrieval will be in {round(INCREASED_INTERVAL / 60, 2)} minutes. "
                    f"For changing the interval size, please run cluster.restart_server(). "
                    f"Interval size will be set to {interval_size}"
                )
            finally:
                await asyncio.sleep(interval_size)

    def send_cluster_logs_to_den(self):
        asyncio.run(self.asend_cluster_logs_to_den())



# Source: https://github.com/skypilot-org/skypilot/blob/feb52cf/sky/utils/command_runner.py

"""Runner for commands to be executed on the cluster."""
import enum
import hashlib
import os
import pathlib
import shlex
import time
from typing import List, Optional, Tuple, Union


# from sky.skylet import log_lib

from runhouse.resources.hardware.sky import (
    constants,
    common_utils,
    log_lib,
    subprocess_utils,
)

##### RH modification #####
import logging
from runhouse.logger import logger

##### RH modification #####


# The git exclude file to support.
GIT_EXCLUDE = '.git/info/exclude'
# Rsync options
RSYNC_DISPLAY_OPTION = '-Pavz'
# Legend
#   dir-merge: ignore file can appear in any subdir, applies to that
#     subdir downwards
# Note that "-" is mandatory for rsync and means all patterns in the ignore
# files are treated as *exclude* patterns.  Non-exclude patterns, e.g., "!
# do_not_exclude" doesn't work, even though git allows it.
RSYNC_FILTER_OPTION = '--filter=\'dir-merge,- .gitignore\''
RSYNC_EXCLUDE_OPTION = '--exclude-from={}'

_HASH_MAX_LENGTH = 10


def _ssh_control_path(ssh_control_filename: Optional[str]) -> Optional[str]:
    """Returns a temporary path to be used as the ssh control path."""
    if ssh_control_filename is None:
        return None
    user_hash = common_utils.get_user_hash()
    path = f'/tmp/skypilot_ssh_{user_hash}/{ssh_control_filename}'
    os.makedirs(path, exist_ok=True)
    return path


def ssh_options_list(
    ssh_private_key: Optional[str],
    ssh_control_name: Optional[str],
    *,
    ssh_proxy_command: Optional[str] = None,
    docker_ssh_proxy_command: Optional[str] = None,
    timeout: int = 30,
    port: int = 22,
    disable_control_master: Optional[bool] = False,
) -> List[str]:
    """Returns a list of sane options for 'ssh'."""
    # Forked from Ray SSHOptions:
    # https://github.com/ray-project/ray/blob/master/python/ray/autoscaler/_private/command_runner.py
    arg_dict = {
        # SSH port
        'Port': port,
        # Supresses initial fingerprint verification.
        'StrictHostKeyChecking': 'no',
        # SSH IP and fingerprint pairs no longer added to known_hosts.
        # This is to remove a 'REMOTE HOST IDENTIFICATION HAS CHANGED'
        # warning if a new node has the same IP as a previously
        # deleted node, because the fingerprints will not match in
        # that case.
        'UserKnownHostsFile': os.devnull,
        # Try fewer extraneous key pairs.
        'IdentitiesOnly': 'yes',
        # Abort if port forwarding fails (instead of just printing to
        # stderr).
        'ExitOnForwardFailure': 'yes',
        # Quickly kill the connection if network connection breaks (as
        # opposed to hanging/blocking).
        'ServerAliveInterval': 5,
        'ServerAliveCountMax': 3,
        # ConnectTimeout.
        'ConnectTimeout': f'{timeout}s',
        # Agent forwarding for git.
        'ForwardAgent': 'yes',
    }
    # SSH Control will have a severe delay when using docker_ssh_proxy_command.
    # TODO(tian): Investigate why.
    # We also do not use ControlMaster when we use `kubectl port-forward`
    # to access Kubernetes pods over SSH+Proxycommand. This is because the
    # process running ProxyCommand is kept running as long as the ssh session
    # is running and the ControlMaster keeps the session, which results in
    # 'ControlPersist' number of seconds delay per ssh commands ran.
    if (ssh_control_name is not None and docker_ssh_proxy_command is None and
            not disable_control_master):
        arg_dict.update({
            # Control path: important optimization as we do multiple ssh in one
            # sky.launch().
            'ControlMaster': 'auto',
            'ControlPath': f'{_ssh_control_path(ssh_control_name)}/%C',
            'ControlPersist': '300s',
        })
    ssh_key_option = [
        '-i',
        ssh_private_key,
    ] if ssh_private_key is not None else []

    if docker_ssh_proxy_command is not None:
        logger.debug(f'--- Docker SSH Proxy: {docker_ssh_proxy_command} ---')
        arg_dict.update({
            'ProxyCommand': shlex.quote(docker_ssh_proxy_command),
        })

    if ssh_proxy_command is not None:
        logger.debug(f'--- Proxy: {ssh_proxy_command} ---')
        arg_dict.update({
            # Due to how log_lib.run_with_log() works (using shell=True) we
            # must quote this value.
            'ProxyCommand': shlex.quote(ssh_proxy_command),
        })

    return ssh_key_option + [
        x for y in (['-o', f'{k}={v}']
                    for k, v in arg_dict.items()
                    if v is not None) for x in y
    ]


class SshMode(enum.Enum):
    """Enum for SSH mode."""
    # Do not allocating pseudo-tty to avoid user input corrupting outputs.
    NON_INTERACTIVE = 0
    # Allocate a pseudo-tty, quit the ssh session after the cmd finishes.
    # Be careful of this mode, as ctrl-c will be passed to remote process.
    INTERACTIVE = 1
    # Allocate a pseudo-tty and log into the ssh session.
    LOGIN = 2


class SSHCommandRunner:
    """Runner for SSH commands."""

    def __init__(
        self,
        ip: str,
        ssh_user: str,
        ssh_private_key: str,
        ssh_control_name: Optional[str] = '__default__',
        ssh_proxy_command: Optional[str] = None,
        port: int = 22,
        docker_user: Optional[str] = None,
        disable_control_master: Optional[bool] = False,
    ):
        """Initialize SSHCommandRunner.

        Example Usage:
            runner = SSHCommandRunner(ip, ssh_user, ssh_private_key)
            runner.run('ls -l', mode=SshMode.NON_INTERACTIVE)
            runner.rsync(source, target, up=True)

        Args:
            ip: The IP address of the remote machine.
            ssh_private_key: The path to the private key to use for ssh.
            ssh_user: The user to use for ssh.
            ssh_control_name: The files name of the ssh_control to use. This is
                used to avoid confliction between clusters for creating ssh
                control files. It can simply be the cluster_name or any name
                that can distinguish between clusters.
            ssh_proxy_command: Optional, the value to pass to '-o
                ProxyCommand'. Useful for communicating with clusters without
                public IPs using a "jump server".
            port: The port to use for ssh.
            docker_user: The docker user to use for ssh. If specified, the
                command will be run inside a docker container which have a ssh
                server running at port DEFAULT_DOCKER_PORT
            disable_control_master: bool; specifies either or not the ssh
                command will utilize ControlMaster. We currently disable
                it for k8s instance.
        """
        self.ssh_private_key = ssh_private_key
        self.ssh_control_name = (
            None if ssh_control_name is None else hashlib.md5(
                ssh_control_name.encode()).hexdigest()[:_HASH_MAX_LENGTH])
        self._ssh_proxy_command = ssh_proxy_command
        self.disable_control_master = disable_control_master
        if docker_user is not None:
            assert port is None or port == 22, (
                f'port must be None or 22 for docker_user, got {port}.')
            # Already checked in resources
            assert ssh_proxy_command is None, (
                'ssh_proxy_command is not supported when using docker.')
            self.ip = 'localhost'
            self.ssh_user = docker_user
            self.port = constants.DEFAULT_DOCKER_PORT
            self._docker_ssh_proxy_command = lambda ssh: ' '.join(
                ssh + ssh_options_list(ssh_private_key, None
                                      ) + ['-W', '%h:%p', f'{ssh_user}@{ip}'])
        else:
            self.ip = ip
            self.ssh_user = ssh_user
            self.port = port
            self._docker_ssh_proxy_command = None

    @staticmethod
    def make_runner_list(
        ip_list: List[str],
        ssh_user: str,
        ssh_private_key: str,
        ssh_control_name: Optional[str] = None,
        ssh_proxy_command: Optional[str] = None,
        disable_control_master: Optional[bool] = False,
        port_list: Optional[List[int]] = None,
        docker_user: Optional[str] = None,
    ) -> List['SSHCommandRunner']:
        """Helper function for creating runners with the same ssh credentials"""
        if not port_list:
            port_list = [22] * len(ip_list)
        return [
            SSHCommandRunner(ip, ssh_user, ssh_private_key, ssh_control_name,
                             ssh_proxy_command, port, docker_user,
                             disable_control_master)
            for ip, port in zip(ip_list, port_list)
        ]

    def _ssh_base_command(self, *, ssh_mode: SshMode,
                          port_forward: Optional[List[int]]) -> List[str]:
        ssh = ['ssh']
        if ssh_mode == SshMode.NON_INTERACTIVE:
            # Disable pseudo-terminal allocation. Otherwise, the output of
            # ssh will be corrupted by the user's input.
            ssh += ['-T']
        else:
            # Force pseudo-terminal allocation for interactive/login mode.
            ssh += ['-tt']
        if port_forward is not None:
            for port in port_forward:
                local = remote = port
                logger.info(
                    f'Forwarding port {local} to port {remote} on localhost.')
                ssh += ['-L', f'{remote}:localhost:{local}']
        if self._docker_ssh_proxy_command is not None:
            docker_ssh_proxy_command = self._docker_ssh_proxy_command(ssh)
        else:
            docker_ssh_proxy_command = None
        return ssh + ssh_options_list(
            self.ssh_private_key,
            self.ssh_control_name,
            ssh_proxy_command=self._ssh_proxy_command,
            docker_ssh_proxy_command=docker_ssh_proxy_command,
            port=self.port,
            disable_control_master=self.disable_control_master) + [
                f'{self.ssh_user}@{self.ip}'
            ]

    def run(
            self,
            cmd: Union[str, List[str]],
            *,
            require_outputs: bool = False,
            port_forward: Optional[List[int]] = None,
            # Advanced options.
            log_path: str = os.devnull,
            # If False, do not redirect stdout/stderr to optimize performance.
            process_stream: bool = True,
            stream_logs: bool = True,
            ssh_mode: SshMode = SshMode.NON_INTERACTIVE,
            separate_stderr: bool = False,
            **kwargs) -> Union[int, Tuple[int, str, str]]:
        """Uses 'ssh' to run 'cmd' on a node with ip.

        Args:
            ip: The IP address of the node.
            cmd: The command to run.
            port_forward: A list of ports to forward from the localhost to the
            remote host.

            Advanced options:

            require_outputs: Whether to return the stdout/stderr of the command.
            log_path: Redirect stdout/stderr to the log_path.
            stream_logs: Stream logs to the stdout/stderr.
            check: Check the success of the command.
            ssh_mode: The mode to use for ssh.
                See SSHMode for more details.
            separate_stderr: Whether to separate stderr from stdout.


        Returns:
            returncode
            or
            A tuple of (returncode, stdout, stderr).
        """
        base_ssh_command = self._ssh_base_command(ssh_mode=ssh_mode,
                                                  port_forward=port_forward)
        if ssh_mode == SshMode.LOGIN:
            assert isinstance(cmd, list), 'cmd must be a list for login mode.'
            command = base_ssh_command + cmd
            proc = subprocess_utils.run(command, shell=False, check=False)
            return proc.returncode, '', ''
        if isinstance(cmd, list):
            cmd = ' '.join(cmd)

        log_dir = os.path.expanduser(os.path.dirname(log_path))
        os.makedirs(log_dir, exist_ok=True)
        # We need this to correctly run the cmd, and get the output.
        command = [
            'bash',
            '--login',
            '-c',
            # Need this `-i` option to make sure `source ~/.bashrc` work.
            '-i',
        ]

        command += [
            shlex.quote(f'true && source ~/.bashrc && export OMP_NUM_THREADS=1 '
                        f'PYTHONWARNINGS=ignore && ({cmd})'),
        ]
        if not separate_stderr:
            command.append('2>&1')
        if not process_stream and ssh_mode == SshMode.NON_INTERACTIVE:
            command += [
                # A hack to remove the following bash warnings (twice):
                #  bash: cannot set terminal process group
                #  bash: no job control in this shell
                '| stdbuf -o0 tail -n +5',
                # This is required to make sure the executor of command can get
                # correct returncode, since linux pipe is used.
                '; exit ${PIPESTATUS[0]}'
            ]

        command_str = ' '.join(command)
        command = base_ssh_command + [shlex.quote(command_str)]

        executable = None
        if not process_stream:
            if stream_logs:
                command += [
                    f'| tee {log_path}',
                    # This also requires the executor to be '/bin/bash' instead
                    # of the default '/bin/sh'.
                    '; exit ${PIPESTATUS[0]}'
                ]
            else:
                command += [f'> {log_path}']
            executable = '/bin/bash'
        return log_lib.run_with_log(' '.join(command),
                                    log_path,
                                    require_outputs=require_outputs,
                                    stream_logs=stream_logs,
                                    process_stream=process_stream,
                                    shell=True,
                                    executable=executable,
                                    **kwargs)

    def rsync(
        self,
        source: str,
        target: str,
        *,
        up: bool,
        # Advanced options.
        log_path: str = os.devnull,
        stream_logs: bool = True,
        max_retry: int = 1,
    ) -> None:
        """Uses 'rsync' to sync 'source' to 'target'.

        Args:
            source: The source path.
            target: The target path.
            up: The direction of the sync, True for local to cluster, False
              for cluster to local.
            log_path: Redirect stdout/stderr to the log_path.
            stream_logs: Stream logs to the stdout/stderr.
            max_retry: The maximum number of retries for the rsync command.
              This value should be non-negative.

        Raises:
            exceptions.CommandError: rsync command failed.
        """
        # Build command.
        # TODO(zhwu): This will print a per-file progress bar (with -P),
        # shooting a lot of messages to the output. --info=progress2 is used
        # to get a total progress bar, but it requires rsync>=3.1.0 and Mac
        # OS has a default rsync==2.6.9 (16 years old).
        rsync_command = ['rsync', RSYNC_DISPLAY_OPTION]

        # --filter
        rsync_command.append(RSYNC_FILTER_OPTION)

        if up:
            # The source is a local path, so we need to resolve it.
            # --exclude-from
            resolved_source = pathlib.Path(source).expanduser().resolve()
            if (resolved_source / GIT_EXCLUDE).exists():
                # Ensure file exists; otherwise, rsync will error out.
                #
                # We shlex.quote() because the path may contain spaces:
                #   'my dir/.git/info/exclude'
                # Without quoting rsync fails.
                rsync_command.append(
                    RSYNC_EXCLUDE_OPTION.format(
                        shlex.quote(str(resolved_source / GIT_EXCLUDE))))

        if self._docker_ssh_proxy_command is not None:
            docker_ssh_proxy_command = self._docker_ssh_proxy_command(['ssh'])
        else:
            docker_ssh_proxy_command = None
        ssh_options = ' '.join(
            ssh_options_list(
                self.ssh_private_key,
                self.ssh_control_name,
                ssh_proxy_command=self._ssh_proxy_command,
                docker_ssh_proxy_command=docker_ssh_proxy_command,
                port=self.port,
                disable_control_master=self.disable_control_master))
        rsync_command.append(f'-e "ssh {ssh_options}"')
        # To support spaces in the path, we need to quote source and target.
        # rsync doesn't support '~' in a quoted local path, but it is ok to
        # have '~' in a quoted remote path.
        if up:
            full_source_str = str(resolved_source)
            if resolved_source.is_dir():
                full_source_str = os.path.join(full_source_str, '')
            rsync_command.extend([
                f'{full_source_str!r}',
                f'{self.ssh_user}@{self.ip}:{target!r}',
            ])
        else:
            rsync_command.extend([
                f'{self.ssh_user}@{self.ip}:{source!r}',
                f'{os.path.expanduser(target)!r}',
            ])
        command = ' '.join(rsync_command)

        backoff = common_utils.Backoff(initial_backoff=5, max_backoff_factor=5)
        while max_retry >= 0:
            returncode, _, stderr = log_lib.run_with_log(
                command,
                log_path=log_path,
                stream_logs=stream_logs,
                shell=True,
                require_outputs=True)
            if returncode == 0:
                break
            max_retry -= 1
            time.sleep(backoff.current_backoff())

        direction = 'up' if up else 'down'
        error_msg = (f'Failed to rsync {direction}: {source} -> {target}. '
                     'Ensure that the network is stable, then retry.')
        subprocess_utils.handle_returncode(returncode,
                                           command,
                                           error_msg,
                                           stderr=stderr,
                                           stream_logs=stream_logs)
